import pandas as pd
import numpy as np
from scipy import stats
import datetime
import os
import math
import pickle
import matplotlib.pyplot as plt 

import sklearn.metrics
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import SGDClassifier
from sklearn.utils import resample
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import LinearSVC
from sklearn.model_selection import cross_val_score

class HumaiProdCategoryFeatureGen:

    def __init__(self):
        self.vectorizer = None
        return
    
    def init(self, df_prod, item_name='item_name', category='category', verbose=True):

        # Generate Term Frequency-Inverse Document Frequency vectors from all item names
        self.vectorizer = TfidfVectorizer(sublinear_tf=True, min_df=5, norm='l2', encoding='latin-1', 
                             ngram_range=(1, 2))
        self.vectorizer = self.vectorizer.fit(df_prod[item_name].tolist())

    def generate(self, item_names):
        if self.vectorizer is None:
            return None

        features = self.vectorizer.transform(item_names)
        print("Words in vocabulary: {0}".format(len(self.vectorizer.vocabulary_)))

        return features

    def save(self, filename):
        pickle.dump( self.vectorizer, open(filename, "wb") )
        return
    
    def load(self, filename):
        self.vectorizer = pickle.load(open(filename, "rb"))
        return


class HumaiProdCategoryModel:

    def __init__(self):
        self.best_model = None
        self.clf = None
        return
    
    def learn(self, df_train, feature_gen, 
              item_name = 'item_name', category = 'category', 
              k_folds=5, 
              balancing=True, balance_thresholds = [5, 300], 
              verbose=True,
              show_plots=False):

        dfg_train = df_train.groupby(category)

        # Balance the data by undersampling classes that exceed a certain size,
        # and oversampling classes that are not big enough.
        item_names = []
        item_categories = []
        if balancing:
            for _, df_group in dfg_train:
                #print(cat)
                if len(df_group) > balance_thresholds[1]:
                    names, categories = resample(df_group[item_name].tolist(), df_group[category].tolist(), 
                                                replace=False, n_samples=balance_thresholds[1], random_state=0)
                elif len(df_group) < balance_thresholds[0]:
                    names, categories = resample(df_group[item_name].tolist(), df_group[category].tolist(), 
                                                replace=True, n_samples=balance_thresholds[0], random_state=0)

                else:
                    names = df_group[item_name].tolist()
                    categories = df_group[category].tolist()

                item_names.extend(names)
                item_categories.extend(categories)
        else:
            item_names = df_train[item_name]
            item_categories = df_train[category]

        if show_plots:
            # Plot balanced classes
            df_balanced = pd.DataFrame()
            df_balanced[item_name] = item_names
            df_balanced[category] = item_categories
            fig = plt.figure(figsize=(20,6))
            df_balanced.groupby(category)[item_name].count().plot.bar(ylim=0)
            plt.show()

        # Create training data
        features = feature_gen.generate(item_names)
        labels = item_categories

        if verbose:
            print("Features shape {0}".format(features.shape))

        # Evaluate different models
        models = {
            'rf':RandomForestClassifier(n_estimators=200, max_depth=3, random_state=0),
            'svc':LinearSVC(),
            'nb':MultinomialNB(),
            'log':LogisticRegression(random_state=0),
        }

        folds = k_folds
        cv_df = pd.DataFrame(index=range(folds * len(models)))
        entries = []
        for name, model in models.items():
            model_name = name#model.__class__.__name__
            accuracies = cross_val_score(model, features, labels, scoring='accuracy', cv=folds)
            for fold_idx, accuracy in enumerate(accuracies):
                entries.append((model_name, fold_idx, accuracy))

        cv_df = pd.DataFrame(entries, columns=['model_name', 'fold_idx', 'accuracy'])

        accs = cv_df.groupby('model_name')['accuracy'].mean()
        self.best_model = accs.idxmax()

        weights = sklearn.utils.class_weight.compute_sample_weight(class_weight='balanced', y=labels, indices=None)
        self.clf = models[self.best_model].fit(features, labels, sample_weight=weights)

        if verbose:
            print("Best model: {0} (accuracy={1})".format(self.best_model, accs.max()))

        return self.best_model, cv_df

    def predict(self, features):
        if self.clf is None:
            return None

        pred = self.clf.predict(features)

        return pred

    def save(self, filename):
        pickle.dump( [self.best_model, self.clf], open(filename, "wb") )
        return
    
    def load(self, filename):
        self.best_model, self.clf = pickle.load(open(filename, "rb"))
        return


import numpy as np
import pandas as pd
import datetime

from random import randrange
from datetime import timedelta

def get_training_features_for_small_data_transactions(
    df_trans,
    chars_transform_pipeline,
    df_trans_chars=None,
    cid_column='Customer_ID',
    datetime_column='Datetime'
):
    
    """
    Return labeled training data for small dataset where most customers place only one
    order.
    
    Training data built as follows:
      observed=0:
        People in order segment 1
        People in order segment 2 and up
      observed=1:
        People in order segment 2 and up, but with their last order removed (treated 
          as being the observed order), turning them into order segment 1 and up.
    
    Parameters
    ----------
    df_trans : Transactions table
    df_trans_chars : Customer transaction characteristics table calculated from the above.
    chars_transform_pipeline : Customer transaction characteristics pipeline used to get
        df_trans_chars
    cid_column : Name of Customer ID column
    datetime_column : Name of transaction datetime column
    
    Returns
    -------
    A dataframe with labeled ('observed' column) training data.
    """
    
    if df_trans_chars is None:
        df_trans_chars = chars_transform_pipeline.transform(df_trans)

    df = df_trans.copy()
    df.loc[:,'rank'] = df.sort_values(by=[cid_column, datetime_column]). \
        groupby(by=cid_column)[datetime_column].rank(method='min', ascending=False)

    # For each customer, remove their most recent transaction. This will 
    # remove all order-segment 1 customers, so we need to add these back in later.
    # Order segment 2 customers become order segment 1, and these are the customers
    # that we are most interested in comparing against the actual order segment 1; those
    # that did not do a repeat purchase.
    df_trans_os2_and_up_minus_last = df[df['rank'] > 1.0].sort_values(
        by=[cid_column, datetime_column]
    ).copy()    

    print("[get_training_features_for_small_data_transactions] Get characteristics for OS2 and up minus last")
    df_chars_os2_and_up_minus_last = chars_transform_pipeline.transform(df_trans_os2_and_up_minus_last)

    df_chars_os1 = df_trans_chars[df_trans_chars.n_trans == 1].copy()
    df_trans_os2_and_up = df_trans_chars[df_trans_chars.n_trans > 1].copy()

    a = set(df_chars_os1[cid_column].to_numpy())
    b = set(df_chars_os2_and_up_minus_last[cid_column].to_numpy())
    assert len( a & b ) == 0, "{:,} customers are in both order segment 1 as well as 2+.".format(len(a&b))

    df_chars_os1.loc[:,'observed'] = 0
    df_trans_os2_and_up.loc[:,'observed'] = 0
    df_chars_os2_and_up_minus_last.loc[:,'observed'] = 1

    full_to_dur = {a:b for a,b in zip(
        df_trans_chars[cid_column].to_numpy(),
        df_trans_chars['days_between_most_recent_trans'].to_numpy()
    ) }

    df_chars_os2_and_up_minus_last.loc[:,'last_trans_dur_days'] = df_chars_os2_and_up_minus_last[cid_column].apply(
        lambda x: full_to_dur.get(x, 0.0)
    )
    
    def random_within_timeframe(start, end):

        delta = end - start
        if delta.days == 0:
            return timedelta(seconds=0)

        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        return timedelta(seconds=randrange(int_delta))

    max_date = df[datetime_column].max()
    df_chars_os1.loc[:,'last_trans_dur_days'] = df_chars_os1.apply(
        lambda x: random_within_timeframe(x['last_trans_date'], max_date), axis=1)
    df_chars_os1.loc[:,'last_trans_dur_days'] = df_chars_os1['last_trans_dur_days'] / pd.to_timedelta(1, unit='D')
    df_trans_os2_and_up.loc[:,'last_trans_dur_days'] = df_trans_os2_and_up.apply(
        lambda x: random_within_timeframe(x['last_trans_date'], max_date), axis=1)
    df_trans_os2_and_up.loc[:,'last_trans_dur_days'] = df_trans_os2_and_up['last_trans_dur_days'] / pd.to_timedelta(1, unit='D')
        
    df_training_chars = df_chars_os1.append([df_chars_os2_and_up_minus_last, df_trans_os2_and_up], sort=True)
    
    return df_training_chars


def get_training_basic_features_transactions(df, date_column, cid, max_hist_len, min_timespan_wks=3, min_unique_wks = 2, make_weeks_whole = True, verbose=False):
    
    df.loc[:,'week_of_year'] = df.loc[:,date_column].apply(lambda x: x.weekofyear)

    dfg_customers = df.sort_values(by=[cid, date_column]).groupby(by=[cid])
    nc = len(dfg_customers)
    print("Number of customers: {:,}".format(nc))
    
    features = []

    ii = 0

    for cid, g in dfg_customers:

        if ii % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))

        timespan_wks = int((g.iloc[-1,:][date_column] - g.iloc[0,:][date_column]).days / 7) + 1
        if timespan_wks >= min_timespan_wks:
            #print(cid)
            #print(g)
            cutoff_wks = np.random.randint(0, max(0, min(max_hist_len-1, timespan_wks-1))) if timespan_wks > 1 else 0
            
            start_date = g.iloc[0,:][date_column]
            end_date = start_date + datetime.timedelta(weeks=cutoff_wks)
            #print("Start: {}, end {}".format(start_date, end_date))
            timespan_wks = int((end_date - g.iloc[0,:][date_column]).days / 7) + 1
            
            transactions = g[ (g[date_column] >= start_date) & (g[date_column] <= end_date) ].copy()
            if transactions['week_of_year'].nunique() >= min_unique_wks:
                #print(g)

                observed = 1 if g.iloc[-1,:][date_column] == transactions.iloc[-1,:][date_column] else 0
                n_trans = len(transactions) - observed

                transactions.drop_duplicates(subset=['week_of_year'], keep='last', inplace=True)
                act_wks = len(transactions) - observed

                if observed:
                    last_trans_dur_wks = int((transactions.iloc[-1,:][date_column] - transactions.iloc[-2,:][date_column]).days / 7) + 1
                    timespan_wks = int((transactions.iloc[-2,:][date_column] - transactions.iloc[0,:][date_column]).days / 7) + 1
                else:
                    last_trans_dur_wks = int((end_date - transactions.iloc[-1,:][date_column]).days / 7) + 1
                    timespan_wks = int((transactions.iloc[-2,:][date_column] - transactions.iloc[0,:][date_column]).days / 7) + 1

                nonact_wks = max(0, timespan_wks - act_wks)
                #print("start_week {}, end_week {}".format(start_week, end_week))
                #print("observed {}, last_trans_dur {}, n_trans {}, act_wks {}, nonact_wks {}, timespan_wks {}".format(observed, last_trans_dur, n_trans, act_wks, nonact_wks, timespan_wks))

                features.append((cid, observed, transactions.iloc[-2,:][date_column], last_trans_dur_wks, n_trans, act_wks, nonact_wks, timespan_wks))
        
        ii = 1 + ii

    print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))
    
    df_features = pd.DataFrame(
        features, columns=[
            'cid', 'observed', 'last_transaction_date', 'last_trans_dur_wks', 'n_trans', 'act_wks', 'nonact_wks', 'timespan_wks'
        ])

    return df_features

def get_prediction_basic_features_transactions(df, timepoint, max_hist_len = 12,
                                           date_column = 'Datetime', cid_column = 'Customer_ID', 
                                           verbose=False):
    
    df.loc[:,'week_of_year'] = df[date_column].apply(lambda x: x.weekofyear)

    dfg_customers = df.sort_values(by=[cid_column, date_column]).groupby(by=[cid_column])
    nc = len(dfg_customers)
    print("Number of customers: {:,}".format(nc))
    print("Transaction stats: {}".format(dfg_customers['Transaction_ID'].count().describe()))
    
    features = []

    ii = 0

    for cid, g in dfg_customers:

        if ii % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))

        if len(g) >= 1:
            #print("cid {}".format(cid))
            #print("g (n={}) {}".format(len(g),g))
            
            end_date = g.iloc[-1,:][date_column]
            start_date = end_date - datetime.timedelta(weeks=max_hist_len)
            #print("Start: {}, end {}".format(start_date, end_date))
            transactions = g[ (g[date_column] >= start_date) & (g[date_column] <= end_date) ].copy()
            timespan_wks = int((transactions.iloc[-1,:][date_column] - transactions.iloc[0,:][date_column]).days / 7) + 1
            
            #print("Resulting frame length: {}".format(len(transactions)))
            #if len(transactions) > 1:
            #    print("yay")
                        
            #last_trans_dur_wks = int((timepoint - transactions.iloc[-1,:][date_column].tz_convert(None)).days / 7) + 1
            last_trans_dur_wks = int((timepoint - transactions.iloc[-1,:][date_column]).days / 7) + 1
            
            n_trans = len(transactions)
            
            #print("transactions {}".format(transactions))
            
            transactions.drop_duplicates(subset=['week_of_year'], keep='last', inplace=True)
            act_wks = len(transactions)

            nonact_wks = max(0, timespan_wks - act_wks)
            #print("start_week {}, end_week {}".format(start_week, end_week))
            #print("n_trans {}, act_wks {}, nonact_wks {}, timespan_wks {}".format(n_trans, act_wks, nonact_wks, timespan_wks))

            features.append((cid, end_date, last_trans_dur_wks, n_trans, act_wks, nonact_wks, timespan_wks))
        
        ii = 1 + ii

    print("  {:,}/{:,} ({:.2f} %)".format(ii, nc, 100.0*ii/nc))
    
    df_features = pd.DataFrame(
        features, columns=[
            cid_column, 'last_trans_date', 'last_trans_dur_wks', 'n_trans', 'act_wks', 'nonact_wks', 'timespan_wks'
        ])

    return df_features




import pandas as pd
import numpy as np
import datetime

# Clip [a,b) to lie within [s,t)
def clip_range(a, b, s, t):
    if b <= s or a >= t:
        return pd.NaT, pd.NaT, True
    
    return min(max(a, s), t), min(max(b, s), t), False

def get_date_range_for_lifecyclestage(df_lifecycle, stages=['hl'], timeframe=None):

    df = df_lifecycle
    
    wk_cols = []

    for c in list(df):
        if type(c) == str and 'week_' in c:
            wk_cols.append(c)
            
    df_out = df.copy()

    clipped = []
    start_dates = []
    end_dates = []
        
    n = len(df_out)
    for i in range(n):
        row = df_out.iloc[i,:]
        
        if i % 5000 == 0:
            print("  {:,}/{:,} ({:.2f} %)".format(i, n, 100.0*i/n))
        
        #print("[{},{})".format(tf_start_wk, tf_end_wk))
        
        cur_stage = 'no_stage'
        stage_start_wk = -1
        stage_end_wk = -1
        # Half-open [stage_start_wk, stage_end_wk)
        for c in wk_cols:
            if row[c] in stages and cur_stage not in stages:
                stage_start_wk = int(c.split('_')[1])
            elif row[c] not in stages and cur_stage in stages:
                stage_end_wk = int(c.split('_')[1])
            cur_stage = row[c]
            
        # Half open interval
        stage_start_date = row.last_trans_date + datetime.timedelta(days=7*(stage_start_wk-1))
        stage_end_date = row.last_trans_date + datetime.timedelta(days=7*(stage_end_wk-1))
        
        stage_start_date_clipped = stage_start_date
        stage_end_date_clipped = stage_end_date
        clip = False
        if timeframe is not None:
            stage_start_date_clipped, stage_end_date_clipped, clip = clip_range(
                stage_start_date, stage_end_date, 
                timeframe[0], timeframe[1]
            )
        
        #print("[{},{}) clipped to [{},{}) gets [{},{})".format(
        #    stage_start_date, stage_end_date, timeframe[0], timeframe[1], stage_start_date_clipped, stage_end_date_clipped
        #))
        
        clipped.append(clip)
        start_dates.append( stage_start_date_clipped )
        end_dates.append( stage_end_date_clipped )
        
    print("  {:,}/{:,} ({:.2f} %)".format(n, n, 100.0))
    
    #print(start_dates[:10])
        
    stages_str = '_'.join(stages)
    df_out.loc[:,'{}_clipped'.format(stages_str)] = clipped
    df_out.loc[:,'{}_start_date_clipped'.format(stages_str)] = start_dates
    df_out.loc[:,'{}_end_date_clipped'.format(stages_str)] = end_dates
    
    return df_out


def filter_customers_in_lifecycle_stage(
    df_features, 
    df_lifecycle, 
    max_emails_per_customer=3, 
    stages=['hl'], 
    timeframe=None, 
    cid_column='Customer_ID'
):
    
    assert (isinstance(stages, list)), "stages argument should be a list."
    
    # Get active_frame (start and end date) for all customers in lifecycle stage [stage], within a given timeframe
    df_lifecycle_clipped = get_date_range_for_lifecyclestage(df_lifecycle, stages=stages, timeframe=timeframe)
    
    # Remove all customers that didn't make the cut from the lifecycle
    df_lifecycle_clipped = df_lifecycle_clipped.drop(index=df_lifecycle_clipped[df_lifecycle_clipped['hl_clipped']].index)
    df_lifecycle_clipped.reset_index(inplace=True)

    df_to_pred = df_features.copy()
    df_to_pred.set_index(cid_column, inplace=True)
    
    # Keep only those customers in df_features that are in lifecycle frame, put in df_to_pred
    print("Before filtering {:,}".format(len(df_to_pred)))

    indices = df_to_pred.index.intersection(df_lifecycle_clipped[cid_column])
    df_to_pred = df_to_pred.reindex(indices)

    print("After filtering {:,}".format(len(df_to_pred)))
    
    df_lifecycle_clipped.set_index(cid_column, inplace=True)
    
    # Convert start and end dates of active_frame to start date and num emails
    df_to_pred.loc[indices,'pred_start_date'] = df_lifecycle_clipped.loc[indices, 'hl_start_date_clipped']
    df_to_pred.loc[indices,'pred_end_date'] = df_lifecycle_clipped.loc[indices, 'hl_end_date_clipped']
    df_to_pred.loc[indices,'num_emails_to_predict'] = df_lifecycle_clipped.loc[indices, :].apply(
        lambda x: min(max_emails_per_customer, int((x['hl_end_date_clipped'] - x['hl_start_date_clipped']).days / 7 + 1)), 
        axis=1
    )
    
    df_lifecycle_clipped.reset_index(inplace=True)
    df_to_pred.reset_index(inplace=True)
    
    return df_to_pred


import pandas as pd
import warnings
import re

import humailib.utils as hu
import humailib.hasher as hh
from humailib.cloud_tools import GoogleCloudStorage, GoogleBigQuery

class GenerateCustomerID:
    
    def __init__(self):
        self.cid_offset = 0
        
        # personifier_to_cid:
        # {
        #  'Personifier' : str(customer_id), 
        #   ...
        # }
        self.personifier_to_cid = {}

        # Each person can have one or more different personifiers, that need
        # the same Customer_ID
        # person_to_personifiers=
        # {
        #  'Person' : [personifier0, personifier1, ...], 
        #   ...
        # }
        self.person_to_personifiers = {}
        
        # Each personifier could refer to one or more different persons, for example
        # if a person entered their first name slightly differently in two occasions.
        # personifiers_to_person=
        # {
        #  'Personifier' : [person0, person1, ...], 
        #   ...
        # }
        self.personifier_to_persons = {}
        
    def _walk_personifiers(self, personifier, people=[], personifiers=[]):
        
        for person in self.personifier_to_persons[personifier]:
            if person not in people:
                people.append(person)
                for m in self.person_to_personifiers[person]:
                    if m not in personifiers:
                        personifiers.append(m)
                        self._walk_personifiers(m, people=people, personifiers=personifiers)
                
        return people, personifiers
                
        
    def add(
        self, 
        df, 
        personifier_col='Email',
        cols_to_match=[],
        verbose=False
    ):   
        """
        Generate unique Customer IDs for people, based on personifier and
        person information from columns.
        
        Each personifier gets a unique Customer ID, initially. And then personifiers
        which are different but have matching person information are merged, and get 
        assigned the same Customer ID.
        
        """
        print("[GenerateCustomerID::add] transforming columns...")
        df = df[[personifier_col]+[c for c in cols_to_match if c is not personifier_col]].copy()
        
        df.loc[:,personifier_col] = df[personifier_col].str.strip().str.upper()
        
        def get_person(x, cols_to_match):
            person = [
                re.sub(r'[^\w]', '', x[c]) \
                  .replace('_','') \
                  .upper() for c in cols_to_match
            ]
            if person is None or len(person) == 0:
                return ''
            
            for p in person:
                if p == '':
                    return ''
                
            return '_'.join(person)
        
        # We assume that [cols_to_match] is the minimum information required to
        # uniquely identify and merge a person, so if any of this information is 
        # missing, mark it as an empty person, so it does not get merged.
        df.loc[:,'Person'] = df.apply(lambda x: get_person(x, cols_to_match), axis=1)
        
        for i,personifier in enumerate(df[personifier_col].unique()):
            self.personifier_to_cid[personifier] = self.personifier_to_cid.get(personifier, 'S'+str(i+self.cid_offset))
        
        personifier_in_common = {}
        
        print("[GenerateCustomerID::add] adding new people to table...")
        
        n_new_people = 0
        n_new_personifiers = 0
        ii = 0
        nn = len(df)
        for index, row in df.iterrows():

            if ii % 10000 == 0:
                print("  {:,}/{:,} rows ({:.2f}%)".format(
                    ii,
                    nn,
                    100.0*float(ii)/nn
                ))

            person = row['Person']
            
            # Don't attempt to merge empty persons
            if person != '': 
                personifier = row[personifier_col]

                if personifier not in self.personifier_to_persons:
                    n_new_personifiers = 1 + n_new_personifiers
                    self.personifier_to_persons[ personifier ] = [ person ]
                elif person not in self.personifier_to_persons[personifier]:
                    self.personifier_to_persons[ personifier ].append(person)

                if person not in self.person_to_personifiers:
                    n_new_people = 1 + n_new_people
                    self.person_to_personifiers[ person ] = [ personifier ]
                elif personifier not in self.person_to_personifiers[person]:
                    self.person_to_personifiers[ person ].append(personifier)
                
            ii = 1 + ii
                

        print("[GenerateCustomerID::add] merging ids...")
        n_personifiers = 0
        n_merged = 0
        nn = len(self.personifier_to_persons)
        for personifier in self.personifier_to_persons.keys():
            # Gather all people and personifiers that can be 'reached' by shared personifiers
            people, personifiers = self._walk_personifiers(personifier, people=[], personifiers=[])
            
            if n_personifiers % 5000 == 0:
                print("  {:,}/{:,} personifiers ({:.2f}%)".format(
                    n_personifiers,
                    nn,
                    100.0*float(n_personifiers)/nn
                ))
            
            #if 'MATERS' in personifier:
            #    print("------======------")
            #    print(personifier)
            #    print("walked people: {}".format(people))
            #    print("walked personifiers: {}".format(personifiers))
            #    for mm in personifiers:
            #        print("cid {}".format(self.personifier_to_cid[mm]))
        
            cid = self.personifier_to_cid[personifier]
            # Determine oldest Customer_ID
            for m in personifiers:
                contestant = self.personifier_to_cid.get(m,'')
                if contestant != '' and (int(contestant[1:]) < int(cid[1:])):
                    cid = contestant
            
            # Set all these people's personifiers to oldest Customer_ID
            for m in personifiers:
                if self.personifier_to_cid.get(m,'') != cid:
                    n_merged = 1 + n_merged
                    if verbose:# or 'MATERS' in personifier:
                        print("{} '{}' and '{}' belong to same person: {} (old cid='{}', new cid='{}')".format(
                            n_merged,
                            m,
                            personifier,
                            self.personifier_to_persons[m][0],
                            self.personifier_to_cid[m],
                            cid
                        ))                    
                    self.personifier_to_cid[m] = cid
                    
            n_personifiers = 1 + n_personifiers
            
        cids = [value for _,value in self.personifier_to_cid.items()]
        
        print("====================")
        print("[GenerateCustomerID::generate_from_transactions] New people: {:,}, new personifiers (='{}') {:,}".format(n_new_people, personifier_col, n_new_personifiers))
        print("[GenerateCustomerID::generate_from_transactions] Unique personifiers (='{}') with assigned Customer IDs: {:,}, unique Customer IDs {:,}".format(personifier_col, len(self.personifier_to_cid),len(set(cids))))
        print("[GenerateCustomerID::generate_from_transactions] Total personifiers (='{}') in merge-able person table: {:,}, of which merged (but weren't before): {:,}\n".format(personifier_col, n_personifiers, n_merged))
                
        return self.personifier_to_cid
    
    
    def get_ids(self, df, personifier_col='Email'):
    
        """
        """
        return df[personifier_col].apply(lambda x: self.personifier_to_cid.get(x.strip().upper(),''))
    
    
    def add_and_get_ids(
        self, 
        df, 
        personifier_col,
        cols_to_match,
        verbose=False
    ):
        
        """
        """
        self.add(df, personifier_col=personifier_col, cols_to_match=cols_to_match, verbose=verbose)
        return self.get_ids(df, personifier_col)
    
    
    def to_dataframe(self):
        
        """
        """
        if len(self.personifier_to_cid) == 0 or len(self.person_to_personifiers) == 0:
            return None
        
        data = []
        for personifier,cid in self.personifier_to_cid.items():
            if personifier in self.personifier_to_persons:
                for person in self.personifier_to_persons[personifier]:
                    data.append((cid, personifier, person))
            else:
                data.append((cid, personifier, ''))
                
        df = pd.DataFrame(data, columns=['Customer_ID','Personifier','Person'])
        hu.columns_as_str(df, columns=['Customer_ID','Personifier','Person'])

        df.drop_duplicates(inplace=True)
                    
        return df
    
    
    def upload_and_replace(self, dataset_table_name, gbq):
        
        """
        """
        if len(self.personifier_to_cid) == 0 or len(self.person_to_personifiers) == 0:
            return False
        
        if gbq is None:
            gbq = GoogleBigQuery()
            
        df = self.to_dataframe()

        hh.encrypt_columns(df, columns=['Personifier','Person'])
        gbq.upload_and_replace_table(df, dataset_table_name)
        
        return True
        
        
    def load(self, dataset_table_name, gbq, flush_cache=True):
        
        """
        """
        self.__init__()
        
        if gbq is None:
            gbq = GoogleBigQuery()
            
        df_customers = hu.load_table(
            gbq, 
            dataset_table_name, 
            table_type='customer_ids', 
            flush_cache=flush_cache
        )
        if df_customers is None:
            print("[GenerateCustomerID::load] Failed to load customers table '{}'".format(dataset_table_name))
            return False
        
        personifier_col = 'Personifier'
        if 'Email' in df_customers:
            ## Old table format.
            personifier_col = 'Email'
        
        hh.decrypt_columns(df_customers, columns=[personifier_col,'Person'])
        
        self.cid_offset = df_customers.Customer_ID.str[1:].astype(int).max() + 1
        self.personifier_to_cid = {
            personifier:cid for (cid,personifier) in zip(
                df_customers.Customer_ID.to_numpy(),
                df_customers[personifier_col].to_numpy()
            )
        }
        
        self.personifier_to_persons = {}
        self.person_to_personifiers = {}
        for index, row in df_customers.iterrows():

            person = row['Person']
            personifier = row[personifier_col]

            if person != '': 
                                    
                if personifier not in self.personifier_to_persons:
                    self.personifier_to_persons[ personifier ] = [ person ]
                elif person not in self.personifier_to_persons[personifier]:
                    self.personifier_to_persons[ personifier ].append(person)

                if person not in self.person_to_personifiers:
                    self.person_to_personifiers[ person ] = [ personifier ]
                elif personifier not in self.person_to_personifiers[person]:
                    self.person_to_personifiers[ person ].append(personifier)
            
        return True
        
        


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd
import sklearn
from sklearn.model_selection import train_test_split as tts

def get_train_validation_test(df, covariate_columns, observed_column = 'observed', training_frac = 0.6, validation_frac = 0.2):
    training_n = int(np.floor(len(df) * training_frac))
    validation_n = int(np.floor(len(df) * validation_frac))
    test_n = len(df) - training_n - validation_n

    df_train, df_test, df_train_y, df_test_y = tts(df[covariate_columns], 
                                                    df[observed_column],
                                                    train_size = training_n,
                                                    test_size = validation_n+test_n,
                                                    shuffle = True, 
                                                    stratify = df[observed_column])

    df_train[observed_column] = df_train_y
    df_test[observed_column] = df_test_y

    df_validation, df_test, df_validation_y, df_test_y = tts(df_test[covariate_columns], 
                                                    df_test[observed_column],
                                                    train_size = validation_n,
                                                    test_size = test_n,
                                                    shuffle = True, 
                                                    stratify = df_test[observed_column])

    df_validation[observed_column] = df_validation_y
    df_test[observed_column] = df_test_y

    return df_train, df_validation, df_test


def get_balanced_set(df, observed_column = 'observed'):

    num_pos = df.observed.sum()
    num_neg = len(df) - num_pos
    sample_n = min(num_pos, num_neg)

    balanced_training_data = pd.DataFrame()
    balanced_training_data = df[df[observed_column] == 1].sample(n=sample_n, random_state=1)
    balanced_training_data = balanced_training_data.append(df[df[observed_column] != 1].sample(n=sample_n, random_state=2))

    # Shuffle so that the labels occur randomly
    balanced_training_data = balanced_training_data.sample(frac=1)

    return balanced_training_data


def x_sin(x):
    return x * np.sin(x)


def sin_cos(x):
    return pd.DataFrame(dict(a=np.sin(x), b=np.cos(x)), index=x)


def rnn_data(data, time_steps, labels=False):
    """
    creates new data frame based on previous observation
      * example:
        l = [1, 2, 3, 4, 5]
        time_steps = 2
        -> labels == False [[1, 2], [2, 3], [3, 4]]
        -> labels == True [3, 4, 5]
    """
    rnn_df = []
    for i in range(len(data) - time_steps):
        if labels:
            try:
                rnn_df.append(data.iloc[i + time_steps].as_matrix())
            except AttributeError:
                rnn_df.append(data.iloc[i + time_steps])
        else:
            data_ = data.iloc[i: i + time_steps].as_matrix()
            rnn_df.append(data_ if len(data_.shape) > 1 else [[i] for i in data_])

    return np.array(rnn_df, dtype=np.float32)


def split_data(data, val_size=0.1, test_size=0.1):
    """
    splits data to training, validation and testing parts
    """
    ntest = int(round(len(data) * (1 - test_size)))
    nval = int(round(len(data.iloc[:ntest]) * (1 - val_size)))

    df_train, df_val, df_test = data.iloc[:nval], data.iloc[nval:ntest], data.iloc[ntest:]

    return df_train, df_val, df_test


def prepare_data(data, time_steps, labels=False, val_size=0.1, test_size=0.1):
    """
    Given the number of `time_steps` and some data,
    prepares training, validation and test data for an lstm cell.
    """
    df_train, df_val, df_test = split_data(data, val_size, test_size)
    return (rnn_data(df_train, time_steps, labels=labels),
            rnn_data(df_val, time_steps, labels=labels),
            rnn_data(df_test, time_steps, labels=labels))


def load_csvdata(rawdata, time_steps, seperate=False):
    data = rawdata
    if not isinstance(data, pd.DataFrame):
        data = pd.DataFrame(data)

    train_x, val_x, test_x = prepare_data(data['a'] if seperate else data, time_steps)
    train_y, val_y, test_y = prepare_data(data['b'] if seperate else data, time_steps, labels=True)
    return dict(train=train_x, val=val_x, test=test_x), dict(train=train_y, val=val_y, test=test_y)


def generate_data(fct, x, time_steps, seperate=False):
    """generates data with based on a function fct"""
    data = fct(x)
    if not isinstance(data, pd.DataFrame):
        data = pd.DataFrame(data)
    train_x, val_x, test_x = prepare_data(data['a'] if seperate else data, time_steps)
    train_y, val_y, test_y = prepare_data(data['b'] if seperate else data, time_steps, labels=True)
    return dict(train=train_x, val=val_x, test=test_x), dict(train=train_y, val=val_y, test=test_y)



import pandas as pd
import numpy as np
import datetime
import random
from timeit import default_timer as timer

from dateutil.relativedelta import relativedelta

import sklearn
import lifelines
from lifelines import AalenAdditiveFitter
from lifelines import CoxPHFitter


def predict_propensity_to_buy(
    df,
    start_date,
    end_date,
    covariate_columns,
    rm,
    pred_threshold,
    dur_column='last_trans_dur_wks',
    cid_column='Customer_ID',
    batch_size = 1000,
    verbose=False
):
    n = len(df)
    propensities = []
    prediction_week_range = int( (end_date - start_date) / pd.to_timedelta(7, unit='D') )
    ofs = 0
    while ofs < n:
        print("{:,}/{:,} ({:.2f}%)".format(ofs, n, float(ofs)*100.0/n))

        cols = [c for c in covariate_columns if c != dur_column]
        end = min(n, ofs+batch_size)

        rows = df.iloc[ofs:end,:]
        to_pred_data = []
        for i in range(end-ofs):
            row = rows.iloc[i,:]
            delta = (start_date - row['last_trans_date']) / pd.to_timedelta(7, unit='D')
           
            for wk in range(prediction_week_range):
                to_pred_data.append(tuple([row[cid_column], delta+wk] + row[cols].tolist()))

        df_to_pred = pd.DataFrame(to_pred_data, columns=[cid_column,dur_column]+cols)
        rm.predict(df_to_pred, covariate_columns, pred_threshold=pred_threshold)
        
        for cid, df_propensities in df_to_pred.groupby(by=cid_column):
            
            pred_propensities = np.float128(df_propensities['pred_ranking'].to_numpy())

            if verbose:
                print("Propensities {}: {}".format(row[cid_column], pred_propensities))

            propensities.append(tuple([cid] + pred_propensities.tolist()))
            
        ofs = batch_size + ofs
    
    dates=[start_date+pd.to_timedelta(i*7, unit='D') for i in range(0, prediction_week_range) ]
    columns=['propensity_to_buy_{}'.format(date.strftime("%Y%m%d")) for date in dates]
    df_out = pd.DataFrame(propensities, columns=[cid_column]+columns)
    
    return df_out


def predict_lifecycles(
    df, 
    prediction_date,
    num_months_ahead,
    lhood_columns, 
    prior_columns, 
    rm, 
    cph, 
    pred_threshold, 
    dur_column='last_trans_dur_wks',
    cid_column='Customer_ID', 
    batch_size=1000,
    delay_until_max=True, 
    verbose=False
):
    
    n = len(df)
    out_data = []
    
    start_date = datetime.datetime(year=prediction_date.year, month=prediction_date.month, day=prediction_date.day) #
    end_date = start_date + relativedelta(months=num_months_ahead)
    prediction_week_range = int( (end_date - start_date) / pd.to_timedelta(7, unit='D') )
    
    ofs = 0
    print("{:,}/{:,} ({:.2f}%)".format(ofs, n, float(ofs)*100.0/n))
    while ofs < n:

        t0 = timer()
        
        l_cols = [c for c in lhood_columns if c != dur_column]
        p_cols = [c for c in prior_columns if c != dur_column]
        end = min(n, ofs+batch_size)

        rows = df.iloc[ofs:end,:]
        batch_size = end-ofs
        
        ##
        ## Get likelihood.
        
        to_pred_data = []
        for i in range(batch_size):
            row = rows.iloc[i,:]
            within_week_offset, weeks_to_start = np.modf((start_date - row['last_trans_date']) / pd.to_timedelta(7, unit='D'))
            weeks_to_start = int(weeks_to_start)

            weeks = weeks_to_start + prediction_week_range
            for wk in range(weeks):
                #print(f'weeks={weeks}')
                to_pred_data.append(tuple([row[cid_column], within_week_offset+wk] + row[l_cols].tolist()))

        df_to_pred = pd.DataFrame(to_pred_data, columns=[cid_column, dur_column]+l_cols)
        rm.predict(df_to_pred, [dur_column]+l_cols, pred_threshold=pred_threshold)
        
        likelihoods = []
        for cid, df_propensities in df_to_pred.groupby(by=cid_column):
            
            propensities = np.float128(df_propensities['pred_ranking'].to_numpy())

            if verbose:
                print("Propensities {}: {}".format(row[cid_column], pred_propensities))

            #print(f'propensities n={len(propensities)}')
            likelihoods.append(tuple(propensities.tolist()))
            
        ##
        ## Get prior.
        
        to_pred_times = []
        weeks = 0
        wk_offset = 0
        for i in range(batch_size):
            row = rows.iloc[i,:]
            within_week_offset, weeks_to_start = np.modf((start_date - row['last_trans_date']) / pd.to_timedelta(7, unit='D'))
            weeks_to_start = int(weeks_to_start)

            if weeks_to_start + prediction_week_range > weeks:
                weeks = weeks_to_start + prediction_week_range
                wk_offset = within_week_offset
                
        priors = cph.predict_survival_function(
            rows.loc[:,p_cols], 
            times=[wk_offset+wk for wk in range(weeks)],
        )
        
        ##
        ## Calculate lifecycle stages
        
        for i in range(batch_size):
            row = rows.iloc[i,:]

            #print(f'cid {row[cid_column]}')

            prior = np.array( priors.iloc[:,i] )
            likelihood = np.array( likelihoods[i] )
            window = min(len(prior), len(likelihood))
            likelihood = likelihood[:window]
            prior = prior[:window]
            
            ml = max(likelihood)
            if delay_until_max:
                idx = random.choice( [ii for ii,j in enumerate(likelihood) if j==ml] )
                prior[:idx] = 1.0
            
            posterior = (likelihood * prior).tolist()
            if len(posterior) < prediction_week_range:
                print(f'i={i} -> len(posterior) < prediction_week_range, len(prior)={len(prior)}, len(likelihood)={len(likelihood)}')
            sgn = [0]
            sgn.extend(np.sign(np.diff(posterior)))

            # jb - just bought
            # nt - neutral
            # go - growing opportunity
            # hl - highest likelihood
            # dc - declining
            # ls - loss
            hl_ids = [ii+1 for ii,p in enumerate(posterior[1:]) if p==posterior[0]]

            stage = np.array(['nt'] * len(posterior))
            stage = ['go' if s > 0 else m for s,m in zip(sgn,stage)]
            stage = ['dc' if s < 0 else m for s,m in zip(sgn,stage)]
            state = 'na'
            for ii,s in enumerate(stage):
                    
                if state == 'go' and s != 'go':
                    stage[ii] = 'hl'
                    state = 'na'
                    if s == 'nt':
                        state = 'go'

                if state != 'go' and s == 'go':
                    state = 'go'
                    
            stage[0] = 'jb'
            for hl_id in hl_ids:
                stage[hl_id] = 'hl'
                
            hl_ids = [ii for ii,s in enumerate(stage) if (s == 'hl' or s == 'hl+jb')]
            if len(hl_ids) > 0:
                stage = [
                    'ls' if (ii > hl_ids[-1] and ii > row['death_from']) else \
                    m for s,ii,m in zip(sgn,range(0,len(sgn)), stage)
                ]
            else:
                stage = [
                    'ls' if (ii > row['death_from']) else \
                    m for s,ii,m in zip(sgn,range(0,len(sgn)), stage)
                ]
                            
            try:
                past_stages = stage[0:-prediction_week_range]
                current_stage = stage[-prediction_week_range:-prediction_week_range+1][0]
                future_stages = stage[-prediction_week_range+1:]
            except:
                print(f'i={i}, prediction_week_range={prediction_week_range}, stage={stage} (n={len(stage)})')
                
                
            weeks_to_past_hl = np.nan
            for ii,s in enumerate(past_stages[::-1]):
                if s == 'hl':
                    weeks_to_past_hl = ii+1
                    break
            weeks_to_future_hl = np.nan
            for ii,s in enumerate(future_stages):
                if s == 'hl':
                    weeks_to_future_hl = ii+1
                    break
            
            if np.isnan(weeks_to_past_hl) and np.isnan(weeks_to_future_hl):
                weeks_to_next_hl = np.nan
            elif np.isnan(weeks_to_past_hl):
                weeks_to_next_hl = weeks_to_future_hl
            elif np.isnan(weeks_to_future_hl):
                weeks_to_next_hl = -weeks_to_past_hl
            elif weeks_to_past_hl < weeks_to_future_hl:
                weeks_to_next_hl = -weeks_to_past_hl
            else:
                weeks_to_next_hl = weeks_to_future_hl
            
            #print(f'past stages {past_stages}')
            #print(f'cur stage {current_stage}')
            #print(f'future stages {future_stages}')
            
            past_max_posterior = np.nan
            current_posterior = np.nan
            future_max_posterior = np.nan
            
            if len(posterior[0:-prediction_week_range]) > 0:
                past_max_posterior = max(posterior[0:-prediction_week_range])
            current_posterior = posterior[-prediction_week_range:-prediction_week_range+1][0]
            if len(posterior[-prediction_week_range+1:]) > 0:
                future_max_posterior = max(posterior[-prediction_week_range+1:])
            
            peak_in_future = True if (('hl' in future_stages) or ('go' in future_stages)) else False
            out_data.append(tuple([
                row[cid_column],
                start_date,
                row['last_trans_date'],
                max(posterior) > pred_threshold,
                True if ('hl' in past_stages or 'hl+jb' in past_stages) else False,
                past_max_posterior,
                current_stage,
                current_posterior,
                peak_in_future,
                future_max_posterior,
                weeks_to_next_hl,
            ]))
        
        ofs = batch_size + ofs        

        t1 = timer()
        print("{:,}/{:,} ({:.2f}%), last batch took {}".format(ofs, n, float(ofs)*100.0/n, str(datetime.timedelta(seconds=(t1-t0)))))
        
        
    date_str = start_date.strftime('%Y%m%d')
    df_out = pd.DataFrame(out_data, columns=[cid_column]+[
        'date',
        'last_trans_date',
        'is_high_propensity_customer',
        'peak_in_past','max_past_posterior_propensity',
        'lifecycle_stage_at_date','posterior_propensity_at_date',
        'peak_in_future','max_future_posterior_propensity',
        'weeks_to_next_peak',
    ])
        
    print("{:,}/{:,} ({:.2f}%)".format(i, n, 100))
        
    return df_out



import numpy as np
import pandas as pd
import datetime
import os
import math

import networkx as nx
from humailib.cloud_tools import GoogleCloudStorage

class HumaiGraphBasedRecommender:

    # Hyper-parameters
    also_bought_bias = 0.0
    also_bought_atten = 0.25
    node_weight_atten = 5.0

    def __init__(self):
        self.df_trans = None
        self.df_items = None
        self.G = None
        self.product_ranking = []
        self.custom_weights = {}
        self.cid_column = 'Customer_ID'
        self.product_column = 'Product_ID'

        self.product_ranking_weight = 1.0
        self.bought_together_weight = 1.0
        self.also_bought_weight = 1.0
        self.also_bought_bias = 1.0

    def init(self, df_trans, df_items, product_column = 'Product_ID', cid_column='Customer_ID'):

        self.df_trans = df_trans#.copy()
        self.df_items = df_items#.copy()
        
        # We assume the product column is an integer.
        self.product_column = product_column
        self.cid_column = cid_column

    def train(self, product_ranking_weight=1.0, bought_together_weight=1.0, also_bought_weight=1.0, verbose=False):

        self.product_ranking_weight = product_ranking_weight
        self.bought_together_weight = bought_together_weight
        self.also_bought_weight = also_bought_weight

        self.__calc_product_ranking(verbose)

        self.G = nx.Graph()

        # Add product nodes to the graph, with their popularity
        # rank as node weight.
        for product,rank,w in self.product_ranking:
            w = self.product_ranking_weight * (1.0 - np.exp(-self.node_weight_atten*w))
            #self.G.add_node( pname, {'prod_rank':rank, 'ranking_weight':w} ) # networkx version 1.x
            self.G.add_node( product, prod_rank=rank, ranking_weight=w )
            #if verbose:
            #    print("{0} - rank:{1}, weight:{2}".format(pname, rank, w))

        self.__generate_bought_together_edges(verbose)
        self.__generate_also_bought_edges(verbose)
        self.__calculate_final_proximity(verbose)

        print("Done learning!")
        
    def get_recommendations_product(self, products):
        rset = set()

        for product in products:
            ranked = self.__get_sorted_recommendations_rank(product)
            rset.update([tuple(x) for x in ranked])
        
        return [product for product,_ in sorted(list(rset), key=lambda x: -x[1])]

    def get_most_popular_products(self, n=5):
        
        products = []
        for p in self.product_ranking:
            
            pr = p[0]
            wt = p[2]
            if pr in self.custom_weights:
                
                wt = self.custom_weights[pr] * wt
                if wt > 0.0:
                    products.append(tuple[pr,wt])
                    
            else:    
                products.append(tuple([pr,wt]))
            
        return [p for p,_ in sorted(products, key=lambda x: -x[1])][:n]

    def __get_sorted_recommendations_rank(self, product):

        n_failed = 0
        candidates = []

        if self.G is not None:
            try:
                for neighbour in self.G.neighbors(product):

                    edge = self.G.edges[product,neighbour]
                    if neighbour not in self.custom_weights:
                        candidates.append( 
                            (neighbour, (1.0 - edge['final_proximity']))
                        )
                    elif self.custom_weights[neighbour] > 0.0:
                        # If custom weight for product [neighbour] is 0, it means
                        # [neighbour] is to be excluded from ever being recommended.
                        candidates.append( 
                            (neighbour, (1.0 - edge['final_proximity']) * self.custom_weights[neighbour])
                        )

                    #print("{0} -> {1}".format(neighbour, edge))
            except:
                print("[__get_sorted_recommendations_rank] fail on {}".format(product))
                n_failed = 1 + n_failed

        #print(candidates)

        sorted_candidates = sorted(candidates, key=lambda x: -x[1])
        return [[p,x] for p,x in sorted_candidates]

    def save(self, filename, gcs):

        if self.G is not None:
            if gcs is None:
                gcs = GoogleCloudStorage()
            
            # Save the graph
            gcs.pickle_and_upload([
                self.G,
                self.product_ranking,
                self.custom_weights,
                self.product_ranking_weight,
                self.bought_together_weight,
                self.also_bought_weight,
                self.cid_column,
                self.product_column
            ], filename)

    def load(self, filename, gcs):

        if gcs is None:
            gcs = GoogleCloudStorage()
            
        # For some reason, assigning straight to self.G, etc in here doesn't work
        # and results in self.G being NoneType. So we load into temp variables first.
        G, pr, cws, prw, btw, abw, cid_col, prod_col = gcs.download_and_unpickle(filename)
        print(type(G))
        
        self.G = G
        self.product_ranking = pr
        self.custom_weights = cws
        self.product_ranking_weight = prw
        self.bought_together_weight = btw
        self.also_bought_weight = abw
        self.cid_column = cid_col
        self.product_column = prod_col
        
        print(type(self.G))
        
    # Dictionary of key=product_ID, value=weight in interval: [0,1]
    def add_custom_weights(self, weights, zero_others=False):
        
        self.custom_weights = weights
        
        if zero_others:
            for p in self.product_ranking:
                if p[0] not in self.custom_weights:
                    self.custom_weights[ p[0] ] = 0.0
                    
    def reset_custom_weights(self):
        
        self.custom_weights = {}
        
    def __calculate_final_proximity(self, verbose):

        # Calculate final proximity value for each edge.
        # The lower this value, the more strongly the two connecting
        # nodes (products) are associated with each other.
        n_edges = len(self.G.edges())
        n = 0
        print("Calculating final proximity values (number of edges={:,})...".format(n_edges))
        # Generate final weights
        for product_1, product_2, data in self.G.edges(data=True):
            #print("{0} - {1}".format(product_1, product_2))
            #weight_1 = self.G.node[product_1]['ranking_weight'] # networkx version 1.x
            #weight_2 = self.G.node[product_2]['ranking_weight']
            weight_1 = self.G.nodes[product_1]['ranking_weight']
            weight_2 = self.G.nodes[product_2]['ranking_weight']
            #print("node weights: {0} - {1}".format(weight_1, weight_2))
            min_strength = 0.0001
            bias = 0.0
            #print(data)
            if data.get('bought_together') is not None:
                edge_strength = self.bought_together_weight * data['bought_together']
            elif data.get('also_bought') is not None:
                edge_strength = self.also_bought_weight * self.also_bought_atten * data['also_bought']
                bias = self.also_bought_weight * self.also_bought_bias

            proximity = (1.0 / (edge_strength + min_strength))*(1.0 - weight_1)*(1.0 - weight_2) + bias
            
            #self.G.edge[product_1][product_2]['final_proximity'] = proximity # networkx version 1.x
            self.G.edges[product_1,product_2]['final_proximity'] = proximity 
            print("Proximity between ({},{}) = {:.4f}".format(product_1, product_2, proximity))

            if verbose:
                n = 1 + n
                if n % 5000 == 0:
                    print("{:,}/{:,}".format(n, n_edges))

    def __generate_also_bought_edges(self, verbose):

        dfg_customers = self.df_trans.groupby(by=self.cid_column)
        dfg_transactions = self.df_items.groupby('Transaction_ID')

        # Generate edges between product nodes that have been bought
        # by the same customer, but not necessarily together in the same transaction.
        N = self.df_trans[self.cid_column].nunique()
        n = 0
        transactions_missing = set()
        print("Generating 'also bought' edges (number of unique customers={:,})...".format(N))
        for _, transactions in dfg_customers:
            #num_trans = int( transactions['transaction_rank'].max() )
            num_trans = len(transactions)
            
            #print(N)
            if num_trans > 1:
                #print("Customer_ID: {0} (num_trans={1})".format(customer_id, num_trans))
                for i in range(num_trans-1):
                    
                    trans_id = transactions.iloc[i,:]['Transaction_ID']
                    
                    try:
                        products_1 = dfg_transactions.get_group(trans_id)[self.product_column].values
                        
                        #print("  {0}".format(itemsA))
                        for j in range(i+1, num_trans):
                            trans_id = transactions.iloc[j,:]['Transaction_ID']
                            
                            try:
                                products_2 = dfg_transactions.get_group(trans_id)[self.product_column].values
                                self.__add_edges(products_1, products_2)
                            except:
                                transactions_missing.add(trans_id)
                    except:
                        transactions_missing.add(trans_id)

            if verbose:            
                n = 1 + n
                if n % 5000 == 0:
                    print("{:,}/{:,}".format(n,N))

        num_transactions = self.df_trans['Transaction_ID'].nunique()

        print("num_transactions_lost: {0} out of {1} total ({2}%)".format(
            len(transactions_missing), num_transactions, 
            100.0*len(transactions_missing)/num_transactions))

    def __generate_bought_together_edges(self, verbose):

        dfg_trans = self.df_items.groupby(by=['Transaction_ID'])

        # Generate edges between product nodes that have been
        # bought together in the same transaction. With
        # "bought_together" weight being the number of times
        # they have shown up in the same basket, across all customers.
        n = 0
        N = self.df_items['Transaction_ID'].nunique()
        print("Generating 'bought together' edges (number of transactions = {:,})...".format(N))
        for _, items in dfg_trans:
            
            for i in range(0, len(items)-1):
                product_1 = items.iloc[i,:][self.product_column]
                for j in range(i+1, len(items)):
                    product_2 = items.iloc[j,:][self.product_column]
                    
                    if product_1 != product_2:
                        data = self.G.get_edge_data(product_1, product_2)
                        if data != None:
                            data['bought_together'] = 1 + data['bought_together']
                            #print("({0} - {1}) + 1 ({2})".format(itemA, itemB, G.edge[itemA][itemB]['same_transaction'])) 
                        else:
                            self.G.add_edge(product_1, product_2, bought_together=1)

            if verbose:
                n = 1 + n    
                if n % 5000 == 0:
                    print("{:,}/{:,}".format(n, N))   

    def __calc_product_ranking(self, verbose):

        if self.df_trans is None or self.df_items is None:
            raise Exception("Attempting to call __calc_product_ranking() when initWithData() has not been called.")

        N = len(self.df_items)
        n = 0

        print("Calculating product popularity ranking (number of line items={:,})...".format(N))

        # Calculate product popularity. 
        # For each product, calculate the number of times it has been purchased.
        products = {}
        for _, item in self.df_items.iterrows():
            #print(data.Item_Name)
            product = item[self.product_column]
            if product not in products:
                #print(key)
                products[product] = 1
            else:
                products[product] = 1 + products[product]

            if verbose:  
                n = 1 + n
                if n % 10000 == 0:
                    print("{:,}/{:,}".format(n,N))

        len(products.keys()) == self.df_items[self.product_column].nunique()

        popular_products = sorted(products.items(), key=lambda x: x[1])

        total_prod_n = len(popular_products)

        self.product_ranking = [
            (product, total_prod_n - rank, float(rank)/total_prod_n) for rank,(product,x) in zip(range(total_prod_n),popular_products)
        ]

    def __add_edges(self, products_1, products_2):

        for product_1 in products_1:
            for product_2 in products_2:

                if product_1 != product_2:
                    data = self.G.get_edge_data(product_1, product_2)
                    if data is not None:
                        data['also_bought'] = 1 + data['also_bought']
                        #print("({0} - {1}) + 1 ({2})".format(itemA, itemB, G.edge[itemA][itemB]['same_transaction']))
                    else:
                        self.G.add_edge(itemA, itemB, also_bought=1)
                        #print("({0} - {1})".format(itemA, itemB))
                        
    
def get_product_recommendations(pr, df, cid_column='Customer_ID', product_column='Product_ID', top_n=5, verbose=True):
        
    ii = 0
    nn = df[cid_column].nunique()
        
    dfg = df.groupby(by=[cid_column])
    recoms = []
    for cid, df in dfg:

        if ii % 500 == 0:
            print("{:,}/{:,} ({:.2f}%)".format(ii, nn, 100.0*ii/nn))

        recom = pr.get_recommendations_product(df[product_column].values)
        if len(recom) < top_n:
            recom = recom + ['']*(top_n - len(recom))

        recoms.append(tuple([cid] + recom[:top_n]))
        #print(recoms[-1])
                
        ii = 1 + ii

    columns = [cid_column] + ['product_{}'.format(i) for i in range(1, top_n+1)]
    df_recom = pd.DataFrame(recoms, columns=columns)

    return df_recom

    
def pad_product_recommendations(pr, df_r, df_products, product_column='Product_ID', method='popular'):
    
    df_r_out = df_r.copy()
    
    num_corrected = 0
    prod_recom_columns = []
    for col in df_r.columns:
        if 'product_' in col:
            prod_recom_columns.append(col)
            
    pad_products = []
    if method=='popular' or method is None:
        pad_products = pr.get_most_popular_products(n=len(prod_recom_columns)) 
    
    pids = {r:i for i,r in enumerate(df_products[product_column].values)}
        
    n = len(df_r)
    for i,row in df_r.iterrows():
    
        if i % 5000 == 0:
            print("{:,}/{:,} ({:.2f} %)".format(i, n, 100.0*i/n))

        j = 0
        #print(i)
        for col in prod_recom_columns:
            #print(row[col])
            if row[col] not in pids or row[col] == '':
                #print(pr.product_ranking[-j][0])
                df_r_out.at[i,col] = str(pad_products[j])
                j = 1 + j
                num_corrected = 1 + num_corrected

    print("Done.")
    
    return df_r_out


class Info(dict):
    
    def __init__(self):
        
        return
    
    def init_for_email(
        self, 
        dataset_base, date_ranges, 
        activity, 
        min_hist_len, max_hist_len, 
        activity_column, activity_datetime_column, 
        send_datetime_column,
        cid_column,
        working_dir
    ):
        self['dataset_base'] = dataset_base
        self['dataset_table_base'] = dataset_base + '.mailings_events'
        self['date_ranges'] = date_ranges
        self['activity'] = activity
        self['min_hist_len'] = min_hist_len
        self['max_hist_len'] = max_hist_len
        self['activity_column'] = activity_column
        self['activity_datetime_column'] = activity_datetime_column
        self['send_datetime_column'] = send_datetime_column
        self['cid_column'] = cid_column
        self['working_dir'] = working_dir
        
        self.__init_datasetname()
        self.__init_intermediate_and_training_files_email()
        self.__init_pipeline_files()
        self.__init_model_files()
        
        
    def init_for_transactions(
        self, 
        dataset_base, date_ranges, max_hist_len,
        datetime_column, cid_column,
        working_dir,
        table_name = None,
    ):
        self['dataset_base'] = dataset_base
        if table_name is not None:
            self['dataset_table_base'] = dataset_base + '.' + table_name
        else:
            self['dataset_table_base'] = dataset_base + '.transactions'
        self['date_ranges'] = date_ranges
        self['activity'] = 'transactions'
        self['max_hist_len'] = max_hist_len
        self['datetime_column'] = datetime_column
        self['cid_column'] = cid_column
        self['working_dir'] = working_dir
        
        self.__init_datasetname()
        self.__init_intermediate_and_training_files_transactions()
        self.__init_pipeline_files()
        self.__init_model_files()
        
        
    def init_for_products(
        self, 
        dataset_base, date_ranges, max_hist_len,
        datetime_column, cid_column, product_column,
        working_dir
    ):
        self['dataset_base'] = dataset_base
        self['dataset_table_base'] = dataset_base + '.transactions'
        self['date_ranges'] = date_ranges
        self['activity'] = 'products'
        self['max_hist_len'] = max_hist_len
        self['datetime_column'] = datetime_column
        self['cid_column'] = cid_column
        self['product_column'] = product_column
        self['working_dir'] = working_dir
        
        self.__init_datasetname()
        self.__init_intermediate_and_training_files_products()      
        self.__init_pipeline_files()
        self.__init_model_files_product()
        
        
    def __init_datasetname(self):
        
        self['source_datasets'] = ['{}_{}'.format(self['dataset_table_base'], dr) for dr in self['date_ranges']]
        self['datasetname'] = '_'.join([self['dataset_table_base']]+self['date_ranges']).replace('.', '_')
        

    def __init_intermediate_and_training_files_base(self):
        
        self['hist_file_cached'] = "{}/data/{}_histories_{}.csv".format(
            self['working_dir'], self['datasetname'], self['activity']
        )
        
        """
        
        self['training_features_file'] = "{}/data/{}_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        self['training_file'] = "{}/data/{}_training_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        self['validation_file'] = "{}/data/{}_validation_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        self['test_file'] = "{}/data/{}_test_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        """
        
        self['all_training_features'] = "{}/data/{}_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        self['training_set'] = "{}/data/{}_training_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        self['validation_set'] = "{}/data/{}_validation_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        self['test_set'] = "{}/data/{}_test_features_max_hist_len_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        
    def __init_intermediate_and_training_files_email(self):
        
        self.__init_intermediate_and_training_files_base()
        
        
    def __init_intermediate_and_training_files_transactions(self):
        
        self.__init_intermediate_and_training_files_base()
        
        self['basic_training_features_file'] = "{}/data/{}_basic_training_features_{}_{}.csv".format(
            self['working_dir'], self['datasetname'], self['max_hist_len'], self['activity']
        )
        
        
    def __init_intermediate_and_training_files_products(self):
        
        self['transactions_training_features_file'] = "{}/data/{}_trans_training_features_{}.csv".format(
            self['working_dir'], self['datasetname'], self['activity']
        )
        
        self['items_training_features_file'] = "{}/data/{}_items_training_features_{}.csv".format(
            self['working_dir'], self['datasetname'], self['activity']
        )
        

    def __init_model_files(self):
        
        self['model_file'] = "{}/model.joblib".format(
            self['working_dir']
        )
        
        
    def __init_model_files_product(self):
        
        self['model_file'] = "{}/model.pkl".format(
            self['working_dir']
        )
        
        
    def __init_pipeline_files(self):
        
        self['preprocess_pipeline_file'] = "{}/preprocess_pipeline.joblib".format(
            self['working_dir']
        )
        
        self['preprocess_functions_file'] = "{}/preprocess_functions.pkl".format(
            self['working_dir']
        )


    
        

        


import pandas as pd
import numpy as np
from scipy import stats
import datetime
import os
import math
import pickle
import matplotlib.pyplot as plt 

from datetime import datetime, timedelta

import lifelines
from lifelines import AalenAdditiveFitter
from lifelines import CoxPHFitter

class HumaiPropHazardsFeatureGen:

    def __init__(self):
        self.dataset = None
        return

    def generate(self, df_trans, date_column='Datetime', cid='Email', min_trans=2, verbose=True):

        df_trans['Datetime'] = pd.to_datetime(df_trans['Datetime'], format='%Y-%m-%dT%H:%M:%S.000Z')
        if verbose:
            print('[HumaiProportionalHazardsModel] Date range from {0} to {1}'.format(df_trans['Datetime'].min(), df_trans['Datetime'].max()))

        if verbose:
            print("  Calculating time between transactions...")

        # Define subset if necessary
        customers_unique = df_trans[cid].unique()
        # Set-up dataframe with data
        df_setup = df_trans[df_trans[cid].isin(customers_unique)]
        # Define transaction_number_renter: Rank transactions per renter
        # 1.Sort on renter_number and transaction_end
        # 2.Rank records
        df_setup.sort_values(by=[cid, date_column], inplace=True)
        groups = df_setup.groupby(by=[cid])[date_column]
        df_setup.loc[:,'transaction_number_customer']=groups.rank(method='first')
        # Calculate time_between_transactions by sorting and using diff() function
        df_setup['time_between_transactions'] = df_setup.sort_values(by=[cid,'transaction_number_customer']).loc[:,date_column].diff()

        # Set time_between_transactions as NaT for first transaction
        mask_first_trans = df_setup.transaction_number_customer==1
        tmp = df_setup['time_between_transactions'].where(~mask_first_trans,other=pd.Timedelta(None))
        #tmp = df_setup['time_between_transactions'].where(~mask_first_trans,other=0.0)
        df_setup['time_between_transactions'] = tmp

        # Create column to days between transactions and version as float
        df_setup.loc[:,'days_between_transactions'] = df_setup['time_between_transactions'].apply(lambda x: max(0, x.ceil(freq='D').days))
        df_setup['days_between_transactions_float'] = df_setup['time_between_transactions'].apply(lambda x: max(0, x.days + x.seconds/(24*60*60)))

        if verbose:
            print("  Getting first transaction...")
        # Add column with first transaction
        a = df_setup[df_setup.transaction_number_customer==1].loc[:,[cid,date_column]]
        b = df_setup
        df_setup = b.merge(how='left',right=a, left_on=cid, right_on=cid,suffixes=('','_first'))

        if verbose:
            print("  Discarding customers with fewer than {0} transactions...".format(min_trans))

        # Mask out the NaTs (first transactions)
        #mask = df_setup.transaction_number_renter == 1
        #df_setup.days_between_transactions_float[mask] = 0.0

        # Keep customers who have performed at least 2 transactions (i.e. they are not right-censored)
        grp = df_setup.groupby(cid)
        tmp = grp.filter(lambda x: x['Transaction_ID'].count() >= min_trans)
        grp = tmp.groupby(by=[cid])

        cids = grp.apply(lambda x: x.name).values

        # Boolean (0,1): Whether total number of transactions > 1, i.e. first-time renters = 0, others = 1
        E = grp.transaction_number_customer.apply(lambda x: min(1, x.count()-1))
        if verbose:
            print("  E...")
            E.head(10)

        #rates = grp.days_between_transactions_float.apply(lambda x: (x.count()+1)/x.sum())
        #if verbose:
        #    print("  rates...")
        #    rates.head(10)

        # Average days between transactions
        days_mean = grp.days_between_transactions_float.mean()
        #days_mean = grp.days_between_transactions_float.median()
        if verbose:
            print("  days_mean...")
            days_mean.describe()

        #days_var = grp.days_between_transactions_float.var()
        days_var = grp.days_between_transactions_float.quantile(0.75) - grp.days_between_transactions_float.quantile(0.25)
        days_var.fillna(value=0,inplace=True)
        if verbose:
            print("  days_var...")
            days_var.head(10)

        last_trans_date = grp.Datetime.last()
        if verbose:
            print("  last transaction date...")
            last_trans_date.head(10)

        # Days between last two transactions
        T = grp.days_between_transactions_float.last()
        if verbose:
            print("  T...")
            T.describe()

        # Total number of transactions
        tcount = grp.transaction_number_customer.count()
        if verbose:
            print("  tcounts...")
            tcount.describe()

        tnumitemstotal = grp['Num_Items_Purchased'].sum()
        if verbose:
            print("  tnumitemstotal...")
            tnumitemstotal.describe()

        frequent_customers = pd.Series(data=[0]*len(tcount))
        mask = (tcount > 5)
        frequent_customers[mask.values] = 1
        mask = (tcount > 10)
        frequent_customers[mask.values] = 2

        #dataset = pd.DataFrame({'frequent_customers' : frequent_customers.values, 'days_mean' : days_mean.values, 'days_var' : days_var.values, 'total' : tcounts.values, 'T' : T.values, 'E' : E.values})
        self.dataset = pd.DataFrame({cid : cids,
                                'last_transaction_date' : last_trans_date.values,
                                'frequent_customers' : frequent_customers.values,
                                'tnumitemstotal' : tnumitemstotal,
                                'days_mean' : days_mean,
                                'days_var' : days_var,
                                'tcount' : tcount,
                                'T' : T.values, 
                                'E' : E.values})

        return self.dataset

    def save(self, filename):
        self.dataset.to_csv(filename)
        #pickle.dump([self.dataset], open(filename, "wb"))
        return

    def load(self, filename):
        self.dataset = pd.read_csv(filename)
        self.dataset['last_transaction_date'] = pd.to_datetime(self.dataset['last_transaction_date'], format='%Y-%m-%d %H:%M:%S')
        #self.dataset = pickle.load( open(filename, "rb") )
        return


class HumaiPropHazardsModel:

    def __init(self):
        self.cph = None
        self.features = None
        
    def learn(self, dataset, features = ['frequent_customers', 'tnumitemstotal', 'days_mean', 'days_var'], 
            duration_col='T', event_col='E', step_size=0.25, 
            verbose=True):

        self.features = features.copy()
        ext_features = features.copy()
        ext_features.extend([duration_col, event_col])
    
        self.cph = CoxPHFitter()
        self.cph.fit(dataset[ext_features], duration_col=duration_col, 
                     event_col=event_col, 
                     step_size=0.25, show_progress=verbose)

        if verbose:
            self.cph.print_summary()

    def predict(self, dataset, features = ['frequent_customers', 'tnumitemstotal', 'days_mean', 'days_var'], threshold=0.8):

        if self.cph is None or self.features is None:
            return None

        if features is None:
            features = self.features

        pred = self.cph.predict_survival_function(dataset[features])

        def findFirstIndex(A, p):
            i = 0
            for a in A:
                if a < p:
                    return i
                i = 1 + i

            return len(A)-1

        #out[:][0].index.values[pred[3]]

        #print(pred)

        predictions = [pred.iloc[:,i].index.values[findFirstIndex(pred.iloc[:,i], threshold)] for i in range(0, pred.shape[1])]

        deltas = [timedelta(days=x) for x in predictions]
        dates = dataset['last_transaction_date'] + deltas

        return predictions, dates

    def save(self, filename):
        pickle.dump([self.cph, self.features], open(filename, "wb"))
        return

    def load(self, filename):
        self.cph, self.features = pickle.load( open(filename, "rb") )
        return

    

    

    


        

        




import requests

def CreateWebhookEmail(name, email, newsletter=False):
    request_body = '{{\"firstname\":\"{0}\",\"prefix\":\"{1}\",\"surname\":\"{2}\",\"email\":\"{3}\",\"newsletter\":\"{4}\"}}'.format(name[0], name[1], name[2], email, "yes" if newsletter else "no")
    return request_body

def CreateJSONFromEmailsGroupsSenddays(emails, groups, sendday):
    json = '{\n\"contacts\": [\n'
    
    i = 0
    
    for e,g,s in zip(emails,groups,sendday):
        
        entry = '{{\n\"email\": \"{0}\",\n\"controlgroup\": \"{1}\",\n\"sendday\":\"{2}\"\n}}'.format(e,g,s)
        if i < (len(emails)-1):
            entry = entry + ','
        
        json = json + entry + '\n'
        
        i = 1 + i
    
    json = json + ']\n}'
    
    return json
    
def SendWebhookRequest(url, json, verbose=False):
    response = requests.post(url, data=json)
    if verbose:
        print("Response code: {0}".format(response.status_code))
        print("Response: " + response.text)

    return response.status_code


import os
import re
import shutil
import dill as pickle
import json
import sys

from google.cloud import storage
from google.cloud import bigquery
from google.cloud.exceptions import NotFound
import google.oauth2
import pandas_gbq
import pandas as pd

from pathlib import Path

import warnings

class GoogleBigQuery:
    
    def __init__(
        self, 
        project, 
        key,
        cache_dir='./cache'
    ):
        self.key = key
        self.credentials = google.oauth2.service_account.Credentials.from_service_account_file(self.key)
        self.gbq = bigquery.Client.from_service_account_json(self.key)
        self.project = project
        self.cache_dir = cache_dir
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        return

    def _table_to_filename(self, dataset_table):

        return dataset_table.replace('.','_') + '.csv'
    
    def _table_exists(self, dataset_table):
    
        try:
            self.gbq.get_table(dataset_table)
            return True
        except NotFound:
            return False

    def download_table_to_pandas(
        self, 
        dataset_table, 
        sql=None, 
        datetime_columns=None, 
        datetime_format='%Y-%m-%d %H:%M:%S', 
        deduplicate=True, 
        flush_cache=False,
        use_pandas_gbq=True
    ):

        cache_file = Path(self.cache_dir + '/' + self._table_to_filename(dataset_table))
        if not flush_cache and cache_file.is_file():
            print("[download_table_to_pandas] Retrieving from cache: {}...".format(cache_file))
            df = pd.read_csv(cache_file)
            
            if datetime_columns is not None:
                for dtc in datetime_columns:
                    df.loc[:,dtc] = pd.to_datetime(df[dtc], format=datetime_format, utc=True).dt.tz_convert(None)
        else:
            if sql is None:
                if deduplicate:
                    sql = """
                    SELECT DISTINCT
                    * 
                    FROM 
                    `{0}`
                    """.format(dataset_table)
                else:
                    sql = """
                    SELECT
                    * 
                    FROM 
                    `{0}`
                    """.format(dataset_table)
            
            if not self._table_exists(dataset_table):
                print("[download_table_to_pandas] Project '{}': Table {} does not exist!".format(self.project, dataset_table))
                return None
            
            print("[download_table_to_pandas] Project '{}': Retrieving {} using query {}...".format(self.project, dataset_table, sql))
            if use_pandas_gbq:
                df = pandas_gbq.read_gbq(
                    sql,
                    #progress_bar_type = 'tqdm',
                    dialect='standard',
                    use_bqstorage_api=True,
                    project_id = self.project,
                    credentials = self.credentials
                )
            else:
                df = self.gbq.query(sql).to_dataframe()

            print("[download_table_to_pandas] Writing {} to cache {}...".format(dataset_table, cache_file))
            df.reset_index(drop=True).to_csv(cache_file, index=False)

            if datetime_columns is not None:
                for dtc in datetime_columns:
                    df.loc[:,dtc] = pd.to_datetime(df[dtc], format=datetime_format, utc=True).dt.tz_convert(None)

        print("[download_table_to_pandas] Retrieved {:,} rows".format(len(df)))

        return df
        
    def upload_table(
        self,
        df,
        dataset_table,
        if_exists = 'append'
    ):
        cache_file = Path(self.cache_dir + '/' + self._table_to_filename(dataset_table))

        print("[upload_and_replace_table] Updating cache for {}...".format(dataset_table))
        df.reset_index(drop=True).to_csv(cache_file, index=False)
        #df.to_csv(cache_file, index=False)

        print("[upload_and_replace_table] Project '{}': Uploading {:,} rows to BQ: {}...".format(self.project, len(df), dataset_table))

        pandas_gbq.to_gbq(
            df,
            destination_table = dataset_table,
            project_id = self.project,
            #progress_bar_type = 'tqdm',
            if_exists = if_exists,
            location = 'EU',
            credentials = self.credentials
        )
        
    def upload_and_replace_table(
        self, 
        df, dataset_table
    ):
        return self.upload_table(df, dataset_table, if_exists='replace')


def loadDatasetToPandas(dataset, from_cloud=False, delimiter=',', convertDatetime=True, datetime_format='%Y-%m-%dT%H:%M:%S.000Z', 
    key='../../service_account_keys/humai-sb-7b9db70de787.json'):
    warnings.warn('please use humailib.cloud_tools.GoogleBigQuery.download_table_to_pandas(...)', category=DeprecationWarning, stacklevel=1)
    return


def uploadDataframeToBQ(dataframe, dataset_name, project='humai-sb', key='../../service_account_keys/humai-sb-7b9db70de787.json'):
    warnings.warn('please use humailib.cloud_tools.GoogleBigQuery.upload_and_replace_table(...)', category=DeprecationWarning, stacklevel=1)
    return

    
# Copyright 2017, Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# This code has been modified slightly by William H. de Boer
# 27/03/2019.

"""Helpers for accessing Google Cloud Storage in Python code.
`pickle_and_upload`: Upload a Python object after pickling to
user-specified `bucket_name` and `object_name`.
`download_and_unpickle`: The opposite of `pickle_and_upload`.
For more information:
https://cloud.google.com/storage/
"""

class GoogleCloudStorage:
    
    def __init__(self, key, cache_dir='./cache'):
        self.storage_client = storage.Client.from_service_account_json(key)
        self.cache_dir = cache_dir
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
            print("Created cache directory '{}'".format(self.cache_dir))

        return
   
    def _make_gcs_uri(self, bucket_name, object_name):
        return 'gs://{}/{}'.format(bucket_name, object_name)

    def _split_uri(self, gcs_uri):
        """Splits gs://bucket_name/object_name to (bucket_name, object_name)"""
        pattern = r'gs://([^/]+)/(.+)'
        match = re.match(pattern, gcs_uri)

        bucket_name = match.group(1)
        object_name = match.group(2)

        return bucket_name, object_name
    
    def _get_cache_file(self, filename):
        
        return self.cache_dir + '/' + filename.split('://')[-1].replace('/','_').replace('-','_')

    def get_blob(self, bucket_name, object_name):
    
        bucket = self.storage_client.get_bucket(bucket_name)
        blob = bucket.blob(object_name)
        return blob

    def get_uri_blob(self, gcs_uri):
        bucket_name, object_name = self._split_uri(gcs_uri)
        return self.get_blob(bucket_name, object_name)

    def list_blobs(self, bucket_name, directory=None):

        if directory is None or len(directory) == 0:
            directory = None
            n = 0
        else:
            add = 0 if directory[-1] == '/' else 1
            n = len(directory) + add # added 1 for the '/'
        
        bucket = self.storage_client.get_bucket(bucket_name)
        blobs = bucket.list_blobs(prefix=directory)

        blobs = [blob.name[n:] or '.' for blob in blobs]

        return blobs
    
    def file_exists(self, bucket_name, filepath):
        
        ret = self.list_blobs(bucket_name, directory=filepath)
        if len(ret) == 0:
            return False
        
        return ret[0] == '.'

    def upload_file(self, source_filename, dest_filename):
        """Archives a directory and upload to GCS.
        Returns the object's GCS uri.
        """

        bucket_name, object_name = self._split_uri(dest_filename)
        #print(bucket_name, object_name)
        bucket = self.storage_client.get_bucket(bucket_name)
        blob = bucket.blob(object_name)
        #print(source_filename)
        blob.upload_from_filename(source_filename)

        print('[upload_file] Uploaded {0} to {1}'.format(source_filename, self._make_gcs_uri(bucket_name, object_name)))

        return self._make_gcs_uri(bucket_name, object_name)

    def download_file(self, source_filename, dest_filename):

        bucket_name, object_name = self._split_uri(source_filename)
        blob = self.get_blob(bucket_name, object_name)
        #print(bucket_name, object_name, dest_filename)
        with open(str(dest_filename), "wb") as file_obj:
            ret = blob.download_to_file(file_obj)

        print("[download_file] Downloading {0} to {1}".format(self._make_gcs_uri(bucket_name, object_name), dest_filename))

        return ret

    def upload_csv(
        self, 
        df,
        dest_filename,
        **kwargs,
    ):

        cache_file = self._get_cache_file(dest_filename)
        print("[upload_csv] Writing to cache: {}...".format(cache_file))
        df.reset_index(drop=True).to_csv(cache_file, index=False, **kwargs)
        #df.to_csv(cache_file, index=False)

        print("[upload_csv] Uploading dataframe ({:,} rows) to {}...".format(len(df), dest_filename))
        self.upload_file(cache_file, dest_filename)

        return
    
    def upload_excel(
        self, 
        df,
        dest_filename):

        cache_file = self._get_cache_file(dest_filename)
        print("[upload_excel] Writing to cache: {}...".format(cache_file))
        df.reset_index(drop=True).to_excel(cache_file, index=False)
        #df.to_csv(cache_file, index=False)

        print("[upload_excel] Uploading dataframe ({:,} rows) to {}...".format(len(df), dest_filename))
        self.upload_file(cache_file, dest_filename)

        return
    
    def upload_json(self, data, dest_filename):
        
        cache_file = self._get_cache_file(dest_filename)
        print("[upload_json] Writing to cache: {}...".format(cache_file))
        with open(cache_file, 'w') as json_file:
            json.dump(data, json_file, indent=4)
        
        print("[upload_json] Uploading to {}...".format(dest_filename))
        self.upload_file(cache_file, dest_filename)
        
        return
        
    def download_csv(
        self, 
        source_filename, 
        datetime_columns=None, datetime_format='%Y-%m-%d %H:%M:%S', 
        flush_cache=True,
        **kwargs,
    ):

        cache_file = Path(self._get_cache_file(source_filename))
        if not flush_cache and cache_file.is_file():
            print("[download_csv] Retrieving from cache: {}...".format(cache_file))
        else:
            print("[download_csv] Downloading and caching: {}...".format(source_filename))
            self.download_file(source_filename, cache_file)    

        df = pd.read_csv(cache_file, **kwargs)
        
        if datetime_columns is not None:
            for dtc in datetime_columns:
                df.loc[:,dtc] = pd.to_datetime(df[dtc], format=datetime_format, utc=True).dt.tz_convert(None)
                
        print("[download_csv] Read {:,} rows".format(len(df)))

        return df

    def download_excel(
        self, 
        source_filename, 
        flush_cache=True,
        **kwargs,
    ):

        cache_file = Path(self._get_cache_file(source_filename))
        if not flush_cache and cache_file.is_file():
            print("[download_excel] Retrieving from cache: {}...".format(cache_file))
        else:
            print("[download_excel] Downloading and caching: {}...".format(source_filename))
            self.download_file(source_filename, cache_file)    
            
        """ if use_cache:
            if cache_file.is_file():
                print("[download_excel] Retrieving from cache: {}...".format(cache_file))
            else:
                print("[download_excel] Downloading and caching {}...".format(source_filename))
                self.download_file(source_filename, cache_file)
        else:
            print("[download_excel] Downloading and re-caching {}...".format(source_filename))
            self.download_file(source_filename, cache_file)"""

        df = pd.read_excel(cache_file, engine='openpyxl', **kwargs)

        print("[download_excel] Read {:,} rows".format(len(df)))

        return df
    
    def download_json(
        self, 
        source_filename, 
        flush_cache=True,
        **kwargs,
    ):

        cache_file = Path(self._get_cache_file(source_filename))
        if not flush_cache and cache_file.is_file():
            print("[download_json] Retrieving from cache: {}...".format(cache_file))
        else:
            print("[download_json] Downloading and caching: {}...".format(source_filename))
            self.download_file(source_filename, cache_file)    

        with cache_file.open(mode='r') as json_file:
            data = json.load(json_file, **kwargs)

        return data


    def pickle_and_upload(self, obj, dest_filename):
        """Returns the object's GCS uri."""
        print('Pickling data')
        pickle_str = pickle.dumps(obj)
        
        bucket_name, object_name = self._split_uri(dest_filename)
        bucket = self.storage_client.get_bucket(bucket_name)
        blob = bucket.blob(object_name)
        print('Uploading to {}...'.format(dest_filename))
        blob.upload_from_string(pickle_str)

        return self._make_gcs_uri(bucket_name, object_name)

    def download_and_unpickle(self, source_filename):
        
        bucket_name, object_name = self._split_uri(source_filename)
        blob = self.get_blob(bucket_name, object_name)
        pickle_str = blob.download_as_string()

        print('Downloading and unpickling {}...'.format(source_filename))
        obj = pickle.loads(pickle_str)
        return obj

    def download_uri_and_unpickle(self, gcs_uri):
        bucket_name, object_name = self._split_uri(gcs_uri)
        obj = self.download_and_unpickle(bucket_name, object_name)

        return obj

    def archive_and_upload(self, bucket_name, directory, extension='zip', object_name=None):
        """Archives a directory and upload to GCS.
        Returns the object's GCS uri.
        """
        object_name = object_name or '{}.{}'.format(directory, extension)

        temp_filename = shutil.make_archive('_tmp', extension, directory)

        bucket = self.storage_client.get_bucket(bucket_name)
        blob = bucket.blob(object_name)
        blob.upload_from_filename(temp_filename)

        os.remove(temp_filename)

        return self._make_gcs_uri(bucket_name, object_name)

       



import numpy as np
import pandas as pd
import tensorflow as tf
from matplotlib import pyplot as plt
import pickle
import os

from humailib.hasher import XORHasher

from tensorflow.contrib import learn
from sklearn.metrics import mean_squared_error

from humailib.lstm import lstm_model
#from libraries.data_processing import generate_data
from humailib.cloud_tools import loadDatasetToPandas
from humailib.sequential_model_helper import GetWeekdaySequencesIncNonOpens

class HumaiSequentialTimingModel:

    # ==========
    #   MODEL
    # ==========

    # Parameters
    learning_rate = 0.005
    training_steps = -1
    batch_size = 128
    display_step = batch_size*2
    epochs = 5

    # Network Parameters
    seq_min_len = 2
    seq_max_len = 10 # Sequence max length
    n_hidden = 256 # hidden layer num of features
    n_classes = 8 # days in the week [0,6] and 'non-open' (day 7)

    dropout_keepprob = 0.5
    clip_gradients = False

    def __init__(self, learning_rate=learning_rate, dropout_keepprob=dropout_keepprob):

        self.session = None
        self.learning_rate = learning_rate
        self.dropout_keepprob = dropout_keepprob

    def setNetworkParams(self, seq_min_len=2, seq_max_len=1, n_hidden=256, n_classes=8):

        self.seq_min_len = seq_min_len
        self.seq_max_len = seq_max_len
        self.n_hidden = n_hidden
        self.n_classes = n_classes

    def learn(self, histories, labels, seqlens, val_size=0.1, test_size=0.1, learning_rate=0.005, batch_size=128, epochs=5, clip_gradients=True, save_intermediate=True, verbose=True):

        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.clip_gradients = clip_gradients
        self.epochs = epochs

        train_x, val_x, test_x = self.__splitData(histories, val_size, test_size)
        train_y, val_y, test_y = self.__splitData(labels, val_size, test_size)
        train_sl, val_sl, test_sl = self.__splitData(seqlens, val_size, test_size)

        X = dict(train=train_x, val=val_x, test=test_x)
        y = dict(train=train_y, val=val_y, test=test_y)
        sl = dict(train=train_sl, val=val_sl, test=test_sl)
            
        del histories
        del labels
        del seqlens
        del train_x, val_x, test_x
        del train_y, val_y, test_y
        del train_sl, val_sl, test_sl

        X['train'] = np.reshape(X['train'], (len(X['train']), self.seq_max_len, 1))
        X['val'] = np.reshape(X['val'], (len(X['val']), self.seq_max_len, 1))
        X['test'] = np.reshape(X['test'], (len(X['test']), self.seq_max_len, 1))

        pickle.dump( X, open( "seq_model_learn_data_X.p", "wb" ) )
        pickle.dump( y, open( "seq_model_learn_data_y.p", "wb" ) )
        pickle.dump( sl, open( "seq_model_learn_data_sl.p", "wb") )

        self.training_steps = int( np.floor(len(X['train'])*self.epochs / self.batch_size) )
        if verbose:
            print("training samples n = {0}, validation n = {1}, test n = {2}".format(len(X['train']), len(X['val']), len(X['test'])))
            print("Number of epochs: {0}, batch size: {1}, size of training data: {2} samples".format(self.epochs, self.batch_size, len(X['train'])))
            print("Number of steps required: {0}".format(self.training_steps))

        self.__setupModel()

        # Initialize the variables (i.e. assign their default value)
        init = tf.global_variables_initializer()

        # Start training
        sess = tf.Session()
        if True:

            # Run the initializer
            sess.run(init)

            feeder = dataFeeder(X['train'], y['train'], sl['train'])

            for step in range(1, self.training_steps+1):

                batch_x, batch_y, batch_seqlen = feeder.getNext(self.batch_size)
                #print("n=({0}, {1}, {2})".format(len(batch_x), len(batch_y), len(batch_seqlen)))
                
                # Run optimization op (backprop)
                sess.run(self.optimizer, feed_dict={self.gix: batch_x, 
                                            self.giy: batch_y,
                                            self.seqlen: batch_seqlen, 
                                            self.keep_prob:self.dropout_keepprob})

                if step % self.display_step == 0 or step == 1:
                    # Calculate batch accuracy & loss
                    acc, loss = sess.run([self.accuracy, self.cost], feed_dict={self.gix: batch_x,
                                                        self.giy: batch_y,
                                                        self.seqlen: batch_seqlen, 
                                                        self.keep_prob:1.0})

                    if verbose:
                        print("Train: Step " + str(step) + ", Minibatch Loss= " + \
                            "{:.6f}".format(loss) + ", Training Accuracy= " + \
                            "{:.5f}".format(acc))

                    acc, loss = sess.run([self.accuracy, self.cost], feed_dict={self.gix: X['val'], 
                                                        self.giy: y['val'],
                                                        self.seqlen: sl['val'], 
                                                        self.keep_prob:1.0})

                    if verbose:
                        print("  Validation Loss= " + \
                            "{:.6f}".format(loss) + ", Validation Accuracy= " + \
                            "{:.5f}".format(acc))

            if verbose:
                print("Optimization Finished!")

            # Calculate accuracy
            test_data = X['test']
            test_label = y['test']
            test_seqlen = sl['test']

            accuracy = sess.run(self.accuracy, feed_dict={self.gix: test_data, self.giy: test_label,
                                                self.seqlen: test_seqlen, self.keep_prob:1.0})
                                                
            if verbose:
                print("Testing Accuracy: {}".format(accuracy))

            conf_matrix = sess.run([self.confusion_matrix], feed_dict={self.gix: test_data, self.giy: test_label, self.seqlen: test_seqlen, self.keep_prob:1.0})

            if verbose:
                print("Confusion matrix on test set:\n{0}".format(conf_matrix))

            self.session = sess

        return True, conf_matrix

    def predict(self, sequences, seqlens):

        if self.session is None:
            return False

        sequences = np.reshape(sequences, (len(sequences), self.seq_max_len, 1))
        predictions = self.session.run(self.pred, feed_dict={self.gix: sequences,
                                                self.seqlen: seqlens, 
                                                self.keep_prob:1.0})

        return predictions

    def saveModel(self, modelname, hyperparamname):

        if self.session is None:
            return False

        # Add ops to save and restore all the variables.
        self.saver = tf.train.Saver(save_relative_paths=True)

        # Save graph, variables, and hyper-parameters.

        save_path = self.saver.save(self.session, modelname, write_meta_graph=True)
        print("Model saved in path: '{0}'".format(save_path))
            
        pickle.dump( [self.learning_rate, self.training_steps,
            self.batch_size, self.display_step,
            self.seq_min_len, self.seq_max_len,
            self.n_hidden, self.n_classes,
            self.dropout_keepprob,
            self.clip_gradients], open( hyperparamname, "wb" ) )
        print("Hyper parameters saved in: '{0}'".format(hyperparamname))

        return True

    def loadModel(self, modelname, hyperparamname):

        if self.session is not None:
            self.session.close()

        #self.saver = tr.train.Saver()

        self.session = tf.Session()
        self.loaded_model = tf.train.import_meta_graph(modelname + '.meta')
        self.loaded_model.restore(self.session, tf.train.latest_checkpoint(os.path.dirname(modelname)))

        self.learning_rate, self.training_steps,
        self.batch_size, self.display_step,
        self.seq_min_len, self.seq_max_len,
        self.n_hidden, self.n_classes,
        self.dropout_keepprob,
        self.clip_gradients = pickle.load(open(hyperparamname, "rb"))

        # Restore tensors
        graph = tf.get_default_graph()
   
        self.gix = graph.get_tensor_by_name("gix:0")
        self.giy = graph.get_tensor_by_name("giy:0")
        self.seqlen = graph.get_tensor_by_name("seqlen:0")
        self.keep_prob = graph.get_tensor_by_name("keep_prob:0")
        self.pred = graph.get_tensor_by_name("pred:0")

        return True

    def __splitData(self, data, val_size=0.1, test_size=0.1):
        """
        splits data to training, validation and testing parts
        """
        ntest = int(round(len(data) * (1 - test_size)))
        nval = int(round(len(data[:ntest]) * (1 - val_size)))

        #df_train, df_val, df_test = np.array(data[:nval], dtype=np.float32), np.array(data[nval:ntest], dtype=np.float32), np.array(data[ntest:], dtype=np.float32)
        df_train, df_val, df_test = np.array(data[:nval]), np.array(data[nval:ntest]), np.array(data[ntest:])

        return df_train, df_val, df_test

    def __setupModel(self):

        # ==========
        #   MODEL
        # ==========

        # tf Graph input
        self.gix = tf.placeholder("float", [None, self.seq_max_len, 1], name="gix")
        self.giy = tf.placeholder("float", [None, self.n_classes], name="giy")
        # A placeholder for indicating each sequence length
        self.seqlen = tf.placeholder(tf.int32, [None], name="seqlen")

        self.keep_prob = tf.placeholder("float", shape=None, name="keep_prob")

        # Define weights
        self.weights = {
            'out': tf.Variable(tf.random_normal([self.n_hidden, self.n_classes]), name="weights")
        }
        self.biases = {
            'out': tf.Variable(tf.random_normal([self.n_classes]), name="biases")
        }

        self.pred = self.__dynamicRNN(self.gix, self.seqlen, self.keep_prob, self.weights, self.biases)
        # Give it a name, so we can use it for model restoring
        self.pred = tf.identity(self.pred, name="pred")

        # Define loss and optimizer
        self.cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits_v2(logits=self.pred, labels=self.giy), name="cost")
        if self.clip_gradients == True:
            org_opt = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate)
            self.opt = tf.contrib.estimator.clip_gradients_by_norm(org_opt, clip_norm=1.0)
        else:
            self.opt = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate)
        self.optimizer = self.opt.minimize(self.cost, name="optimizer")

        # Evaluate model
        #correct_pred = tf.equal(tf.argmax(pred,1), tf.argmax(giy,1), name="correct_pred")
        #accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32), name="accuracy")
        self.accuracy = tf.contrib.metrics.accuracy(tf.argmax(self.giy,1), tf.argmax(self.pred,1), name="accuracy")
        self.f1_score = tf.contrib.metrics.f1_score(tf.argmax(self.giy,1), tf.argmax(self.pred,1), name="f1_score")
        self.confusion_matrix = tf.confusion_matrix(tf.argmax(self.giy,1), tf.argmax(self.pred,1), name="confusion_matrix")
    

    def __dynamicRNN(self, x, seqlen, keep_prob, weights, biases):

        # Prepare data shape to match `rnn` function requirements
        # Current data input shape: (batch_size, n_steps, n_input)
        # Required shape: 'n_steps' tensors list of shape (batch_size, n_input)
        
        # Unstack to get a list of 'n_steps' tensors of shape (batch_size, n_input)
        x = tf.unstack(x, self.seq_max_len, 1)

        # Define a lstm cell with tensorflow
        lstm_cell = tf.contrib.rnn.BasicLSTMCell(self.n_hidden)

        #cell = tf.nn.rnn_cell.LSTMCell(state_size, state_is_tuple=True)
        #cell = tf.nn.rnn_cell.DropoutWrapper(cell, output_keep_prob=0.5)
        #cell = tf.nn.rnn_cell.MultiRNNCell([cell] * num_layers, state_is_tuple=True)
        lstm_cell = tf.nn.rnn_cell.DropoutWrapper(lstm_cell, output_keep_prob=keep_prob)

        # Get lstm cell output, providing 'sequence_length' will perform dynamic
        # calculation.
        outputs, _ = tf.contrib.rnn.static_rnn(lstm_cell, x, dtype=tf.float32,
                                    sequence_length=seqlen)

        # When performing dynamic calculation, we must retrieve the last
        # dynamically computed output, i.e., if a sequence length is 10, we need
        # to retrieve the 10th output.
        # However TensorFlow doesn't support advanced indexing yet, so we build
        # a custom op that for each sample in batch size, get its length and
        # get the corresponding relevant output.

        # 'outputs' is a list of output at every timestep, we pack them in a Tensor
        # and change back dimension to [batch_size, n_step, n_input]
        outputs = tf.stack(outputs)
        outputs = tf.transpose(outputs, [1, 0, 2])

        # Hack to build the indexing and retrieve the right output.
        batch_size = tf.shape(outputs)[0]
        # Start indices for each sample
        index = tf.range(0, batch_size) * self.seq_max_len + (seqlen - 1)
        # Indexing
        outputs = tf.gather(tf.reshape(outputs, [-1, self.n_hidden]), index)

        # Linear activation, using outputs computed above
        ppp = 0.99*(tf.matmul(outputs, weights['out']) + biases['out']) + 0.001
        #print(ppp)
        
        return ppp

class dataFeeder:

    def __init__(self, X, y, seqlen):
        self.X = X
        self.y = y
        self.seqlen = seqlen

        self.index = 0

    def getNext(self, batch_size = 128):

        ofs = self.index

        if ofs + batch_size > len(self.X):
            ofs = 0

        X = self.X[ofs:ofs+batch_size]
        y = self.y[ofs:ofs+batch_size]
        seqlen = self.seqlen[ofs:ofs+batch_size]

        self.index = batch_size + self.index

        return X, y, seqlen


    


import pandas as pd
import numpy as np

from pathlib import Path

# Classifier models
import sklearn
import joblib

from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import SGDClassifier, LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import cross_val_score

from sklearn.pipeline import FeatureUnion, Pipeline

from humailib.cloud_tools import GoogleCloudStorage

class HumaiRankingModel:

    def __init__(self):

        self.models = {
            'rf':RandomForestClassifier(n_estimators=200, max_depth=3, random_state=0),
            #'svc':SVC(),  -> slow to train
            'sgd':SGDClassifier(),
            'mlp':MLPClassifier(),
            'nb':MultinomialNB(),
            'log':LogisticRegression(solver='lbfgs', random_state=0),
        }

        self.clf = None

        return

    def select_best_model(self, df, covariate_columns, observed_column = 'observed', scoring='accuracy', folds=5):

        cv_df = pd.DataFrame(index=range(folds * len(self.models)))
        entries = []
        for name, model in self.models.items():
            print("Evaluating {}...".format(name))
            accuracies = cross_val_score(model, df[covariate_columns], df[observed_column], scoring=scoring, cv=folds)
            for fold_idx, accuracy in enumerate(accuracies):
                entries.append((name, fold_idx, accuracy))

        cv_df = pd.DataFrame(entries, columns=['model_name', 'fold_idx', 'accuracy'])

        accs = cv_df.groupby('model_name')['accuracy'].mean()
        print("Best model: {} (mean accuracy: {:.2f})".format(accs.idxmax(), accs.max()))

        return accs.idxmax(), cv_df
    

    def train_model(self, df, covariate_columns, model, observed_column = 'observed'):
        
        self.clf = self.models[model].fit(df[covariate_columns], df[observed_column])
        
        return self.clf
    

    def predict(self, df, covariate_columns, pred_threshold = 0.5, ranking_output = 'pred_ranking', pred_output = 'pred_observed'):

        if self.clf is None:
            return

        df.loc[:,ranking_output] = self.clf.predict_proba(df[covariate_columns])[:,1]
        df.loc[:,pred_output] = df[ranking_output].apply(lambda x: 1 if x > pred_threshold else 0)
        

    def predict_and_rank(self, df, covariate_columns, pred_threshold = 0.5, ranking_output = 'pred_ranking', pred_output = 'pred_observed'):

        if self.clf is None:
            return

        self.predict(df, covariate_columns, pred_threshold=pred_threshold, ranking_output=ranking_output, pred_output=pred_output)
        df.sort_values(by=[ranking_output], ascending=False, inplace=True)
        

    def save(self, filename, gcs=None, cache_dir='./cache'):

        if self.clf is None:
            return

        # Create the model as a single-step pipeline
        pipeline = Pipeline([
            ('classifier', self.clf)
        ])

        if 'gs://' in filename:
            if gcs is None:
                gcs = GoogleCloudStorage()
            cache_file = cache_dir + '/' + filename.split('/')[-1]
            joblib.dump(pipeline, cache_file)
            gcs.upload_file(cache_file, filename)
        else:
            joblib.dump(pipeline, filename)
            

    def load(self, filename, gcs=None, cache_dir='./cache'):

        if 'gs://' in filename:
            if gcs is None:
                gcs = GoogleCloudStorage()
            cache_file = Path(cache_dir + '/' + filename.split('/')[-1])
            gcs.download_file(filename, cache_file)
            pipeline = joblib.load(cache_file)
        else:
            pipeline = joblib.load(filename)
    
        self.clf = pipeline.named_steps['classifier']


    def __test__(self):

        hyper_param_opt = False

        if hyper_param_opt:
            #Hyper parameter optimisation
            # Evaluate different models
            models = {
                'rf1':RandomForestClassifier(n_estimators=200, max_depth=3, random_state=0),
                'rf2':RandomForestClassifier(n_estimators=200, max_depth=2, random_state=0),
                'rf3':RandomForestClassifier(n_estimators=200, max_depth=1, random_state=0),
                'rf4':RandomForestClassifier(n_estimators=200, max_depth=None, random_state=0),
                'rf5':RandomForestClassifier(n_estimators=300, max_depth=None, random_state=0),
                'rf6':RandomForestClassifier(n_estimators=400, max_depth=None, random_state=0),
            }

            folds = 5
            cv_df = pd.DataFrame(index=range(folds * len(models)))
            entries = []
            for name, model in models.items():
                model_name = name#model.__class__.__name__
                accuracies = cross_val_score(model, model_train_feat[covariate_columns], model_labels, scoring='accuracy', cv=folds)
                for fold_idx, accuracy in enumerate(accuracies):
                    entries.append((model_name, fold_idx, accuracy))

            cv_df = pd.DataFrame(entries, columns=['model_name', 'fold_idx', 'accuracy'])

            accs = cv_df.groupby('model_name')['accuracy'].mean()
            best_model = accs.idxmax()
            print(best_model)
            print(cv_df[cv_df.model_name == best_model])
    




"""A simple example of how to access the Google Analytics API."""

import argparse

from apiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials

import httplib2
from oauth2client import client
from oauth2client import file
from oauth2client import tools

class HumaiGAWrapper:

    # Define the auth scopes to request.
    scope = ['https://www.googleapis.com/auth/analytics.readonly']

    def __init__(self):
        self.service = None
        self.profile = None
        self.credentials = None

    def authenticate(self, service_account_email, key_file_location):

        # TODO: Make robust to failure.

        # Authenticate and construct service.
        self.service, self.credentials = self._get_service('analytics', 'v3', self.scope, key_file_location, service_account_email)
        self.profile = self._get_first_profile_id(self.service)
        
        #results = get_results(service, profile)
        accounts = self.service.management().accounts().list().execute()
        print("Username: {0}, Kind: {1}".format(accounts['username'], accounts['kind']));

    def get(self, start_date, end_date, metrics, dimensions, start_index=0, max_rows=1000):
        
        if self.service is None or self.profile is None or self.credentials is None:
            return None

        ids = "ga:" + self.profile
        data = self.service.data().ga().get(
            ids=ids, start_date=start_date, end_date=end_date, metrics=metrics,
            dimensions=dimensions, start_index=start_index, max_results=max_rows).execute()
 
        return data["rows"], data["totalResults"]

    def get_source_sessions(self, start_date, end_date, source_filters):

        if self.service is None or self.profile is None or self.credentials is None:
            return None

        ids = "ga:" + self.profile
        metrics = "ga:sessions"
        data = self.service.data().ga().get(
            ids=ids, start_date=start_date, end_date=end_date, metrics=metrics,
            filters=source_filters).execute()
        #printDictionary(data)
        return data["totalsForAllResults"][metrics]


    def _get_service(self, api_name, api_version, scope, key_file_location,
                    service_account_email):
        """Get a service that communicates to a Google API.

        Args:
        api_name: The name of the api to connect to.
        api_version: The api version to connect to.
        scope: A list auth scopes to authorize for the application.
        key_file_location: The path to a valid service account p12 key file.
        service_account_email: The service account email address.

        Returns:
        A service that is connected to the specified API.
        """

        credentials = ServiceAccountCredentials.from_p12_keyfile(
        service_account_email, key_file_location, scopes=scope)

        http = credentials.authorize(httplib2.Http())
        print(http)

        # Build the service object.
        service = build(api_name, api_version, http=http)

        return service, credentials

    def _get_first_profile_id(self, service):
        # Use the Analytics service object to get the first profile id.

        # Get a list of all Google Analytics accounts for this user
        accounts = service.management().accounts().list().execute()

        if accounts.get('items'):
            # Get the first Google Analytics account.
            account = accounts.get('items')[0].get('id')

            # Get a list of all the properties for the first account.
            properties = service.management().webproperties().list(
                accountId=account).execute()

            if properties.get('items'):
                # Get the first property id.
                property = properties.get('items')[0].get('id')

                # Get a list of all views (profiles) for the first property.
                profiles = service.management().profiles().list(
                accountId=account,
                webPropertyId=property).execute()

                if profiles.get('items'):
                    # return the first view (profile) id.
                    return profiles.get('items')[0].get('id')

        return None



import requests
import rauth
from rauth import OAuth2Service
import json

from timeit import default_timer as timer
import datetime

import numpy as np
import pandas as pd

from humailib.cloud_tools import GoogleBigQuery

import humailib.utils as hu
import humailib.hasher as hh


class CopernicaREST:
    
    def __init__(self, access_token):
        
        self.access_token = access_token
        self.base_url = 'https://api.copernica.com/v2/'
        # Just a random string
        self.state = 'sjDIEj39fkDSUo40'
        
    def get_auth_url(self, client_id = '352576017e08e77410efc978040b7e7d', client_secret = 'dc0ac7e6d01998f78a92850835004d3d'):
        
        copernica = OAuth2Service(
            client_id=client_id,
            client_secret=client_secret,
            name='Copernica',
            authorize_url='https://www.copernica.com/nl/authorize',
            access_token_url='https://api.copernica.com/v2/token',
            base_url='https://www.copernica.com/nl/')

        redirect_uri = 'https://www.copernica.com/nl/'
        params = {'state': self.state,
                  'response_type': 'code',
                  'redirect_uri': redirect_uri}

        url = copernica.get_authorize_url(**params)
        return url
        
    def post(self, endpoint, params=None, data=None, verbose=False):

        url = self.base_url + endpoint + '?access_token={}'.format(self.access_token)
        response = requests.post(url, params=params, data=data)
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {0}".format(response.status_code))
            print("Response: " + response.text)

        return response
    
    def delete(self, endpoint, verbose=False):
        
        url = self.base_url + endpoint + '?access_token={}'.format(self.access_token)
        response = requests.delete(url)
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {0}".format(response.status_code))
            print("Response: " + response.text)

        return response
    
    def put(self, endpoint, params=None, data=None, verbose=False):
        
        url = self.base_url + endpoint + '?access_token={}'.format(self.access_token)
        response = requests.put(url, params=params, data=data)
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {0}".format(response.status_code))
            print("Response: " + response.text)

        return response
    
    def get(self, endpoint, params=None, verbose=False):
        
        url = self.base_url + endpoint + '?access_token={}'.format(self.access_token)
        response = requests.get(url, params=params)
        if verbose:
            print("REST URI: {}".format(url))
            print("Response code: {0}".format(response.status_code))
            print("Response: " + response.text)

        return response, response.json()
        

class CopernicaEmailExtract:
    
    def __init__(self, access_token):
        
        self.api = CopernicaREST(access_token)
        
    def extract(
        self, from_date, to_date, 
        min_destinations = 3, page_size = 1000, verbose=False
    ):
        """
        Get all email events for the date range in the halt-open interval [from_date, to_date).
        """
        
        total = 1
        offset = 0

        mailings = []
        events = []
        
        ii = 1

        while offset < total:

            response, mm = self.api.get(
                'publisher/emailings', params={
                    'type': 'mass', 
                    'start': offset, 'count': page_size,
                    'fromdate': from_date.strftime('%Y-%m-%d'), 'todate': to_date.strftime('%Y-%m-%d')
                }, 
                verbose=False
            )
            total = mm['total']

            for m in mm['data']:
                mailing_id = m['id']
                mailing_datetime = m['timestamp']
                ndestinations = int(m['destinations'])
                if ndestinations >= min_destinations:

                    _, stats = self.api.get('publisher/emailing/{}'.format(mailing_id), verbose=False)
                    
                    broadcast_id = str(stats['id'])
                    #print(stats)
                    print("{}/{} emailing {} at {}, subject '{}', destinations: {:,}, clicks: {:,}, opens: {:,}".format(
                        ii, total,
                        broadcast_id, 
                        stats['timestamp'], 
                        stats['subject'], 
                        int(stats['destinations']),
                        int(stats['clicks']), 
                        int(stats['impressions'])
                    ))

                    mailings.append((broadcast_id, stats['timestamp'], str(stats['document']), 
                                     stats['subject'], int(stats['destinations']), 
                                     int(stats['impressions']), int(stats['clicks'])))

                    ## events
                    dd = self._get_events_by_type(
                        mailing_id, mailing_datetime, type='deliveries', event_name='DELIVERY', verbose=verbose
                    )
                    events.extend(dd)

                    ## OPENS ("impressions")
                    oo = self._get_events_by_type(
                        mailing_id, mailing_datetime, type='impressions', event_name='OPEN', verbose=verbose
                    )
                    events.extend(oo)

                    ## CLICKS
                    cc = self._get_events_by_type(
                        mailing_id, mailing_datetime, type='clicks', event_name='CLICK', verbose=verbose
                    )
                    events.extend(cc)

                    ## UNSUBSCRIBES 
                    uu = self._get_events_by_type(
                        mailing_id, mailing_datetime, type='unsubscribes', event_name='UNSUBSCRIBE', verbose=verbose
                    )
                    events.extend(uu)
                else:
                    print("{}/{} destinations: {:,}. Doesn't satisfy minimum {:,}".format(
                        ii, total,
                        ndestinations,
                        min_destinations
                    ))

                ii = 1 + ii
                
            offset = mm['count'] + offset

        df_mailings = pd.DataFrame(mailings, columns=['Broadcast_ID', 'Send_Datetime', 'Campaign_ID', 
                                                      'Campaign_Name', 'Num_Customers', 'Num_Opens', 'Num_Clicks'])
        df_mailings = df_mailings.sort_values(by=['Send_Datetime'])
        
        df_mailings_events = pd.DataFrame(events, columns=['Broadcast_ID', 'Copernica_ID', 'Send_Datetime', 
                                                          'Activity', 'Activity_Datetime', 'Description'])
        df_mailings_events = df_mailings_events.sort_values(by=['Activity_Datetime'])
        
        hu.columns_as_datetime(df_mailings, columns=['Send_Datetime'])
        hu.columns_as_datetime(df_mailings_events, columns=['Send_Datetime','Activity_Datetime'])

        return df_mailings, df_mailings_events
        
    
    def _get_events_by_type(self, mailing_id, mailing_datetime, type='events', event_name='DELIVERY', verbose=False):
        
        events = []
        
        if verbose:
            print("[CopernicaEmailExtract_get_events_by_type]... retrieving {} for {}".format(event_name, mailing_id))

        start = 0
        count = 1000
        total = 1
        while start < total:
            response, dd = self.api.get(
                'publisher/emailing/{}/{}'.format(mailing_id, type), 
                params={'start': start, 'count':count},
                verbose=False
            )
            
            if 200 <= response.status_code <= 299:
                
                total = dd['total']

                if verbose:
                    print("  Emailing: {}, found {} [{},{}) total {}".format(mailing_id, event_name, dd['start'], dd['start']+dd['count'], total))

                for d in dd['data']:

                    if d['emailing'] != mailing_id:
                        print("{} {}".format(d['emailing'], mailing_id))
                        
                    broadcast_id = str(mailing_id)
                    events.append((
                        # Broadcast_ID, Copernica_ID, Send_datetime, Activity, Activity_Datetime, Description
                        broadcast_id, d['profile'], mailing_datetime, event_name, d['timestamp'], ''
                    ))

                start = dd['count'] + start
            else:
                print("[CopernicaEmailExtract_get_events_by_type] {}".format(response.text))

        print("  Emailing {}, found {:,} events of type {}".format(mailing_id, len(events), event_name))

        return events     
     
    
class CopernicaPredictionsLoader:
    
    def __init__(self, access_token, database, humai_pred_collection = 'HumaiPredictions'):
        
        self.api = CopernicaREST(access_token)
        self.ready = False
        self.humai_database_id = -1
        self.humai_collection_id = -1
        
    def create_or_init(self, database='HumaiTestDB', humai_pred_collection = 'HumaiPredictions'):
        
        self._create_fields_and_collection(database, humai_pred_collection)
        
    def _create_elements(self, endpoint, df):
    
        _, data = self.api.get(endpoint)

        existing_elements = {d['name']:True for d in data['data']}

        success = True
        for index, row in df.iterrows():

            if row['name'] in existing_elements:
                print("[_create_elements] Found {}".format(row['name']))
            else:
                print("[_create_elements] Creating {}...".format(row['name']))

                new_element = row.to_dict()
                response = self.api.post(endpoint, data=new_element, verbose=True)

                if response.status_code == 201:
                    print("  Created.")
                else:
                    print("  Failed to create.")
                    success = False
                    
        return success
        
    def _create_fields_and_collection(self, database, humai_pred_collection = 'Humai_Predictions'):
        
        ## Set data
        
        self.humai_database_id = -1
        self.humai_collection_id = -1
        self.ready = False
        
        required_fields = [
            ('Humai_Customer_ID','text','',True),
            ('Humai_Group','text','',True)
        ]

        df_fields = pd.DataFrame(required_fields, columns=['name','type','value','displayed'])

        required_interests = [
        #    ('OS_1', 'Humai_Interests'),
        #    ('OS_2',' Humai_Interests'),
        #    ('SIX_MONTH', 'Humai_Interests')
        ]

        df_interests = pd.DataFrame(required_interests, columns=['name','group'])

        required_collection_fields = [
            ('Humai_Customer_ID','text','',True),
            ('Humai_Mail_ID','integer',-1,True),
            ('Type','text','PersonalizedMailFlow',True),
            ('Date','date','2020-01-01',True),
            ('DatetimeSlot','text','Midnightto12PM', True),
            ('Group','text','A',True)
        ]

        df_collection_fields = pd.DataFrame(required_collection_fields, columns=['name','type','value','displayed'])
        
        ## Creation logic
        
        response, data = self.api.get('databases')
        if response.status_code == 200:
            
            databases = {db['name']:db['ID'] for db in data['data']}
            self.humai_database_id = int(databases[database]) if database in databases else -1
                
        if self.humai_database_id >= 0:
            
            print("[_create_fields_and_collection] Found database '{}', ID: {}".format(database, self.humai_database_id))
            
            response, data = self.api.get('database/{}/collections'.format(self.humai_database_id))
            if response.status_code != 200:
                print("[_create_fields_and_collection] Could not retrieve collections for database {}".format(database))
                return
                
            collections = {c['name']:c['ID'] for c in data['data']}
            #print(collections)
            self.humai_collection_id = int(collections[humai_pred_collection]) if humai_pred_collection in collections else -1
            #print(self.humai_collection_id)
            
            if self.humai_collection_id == -1:
                
                response = self.api.post('database/{}/collections'.format(self.humai_database_id), data={'name':humai_pred_collection})
                if response.status_code != 201:
                    print("Could not create collection: {}".format(humai_pred_collection))
                    return
                
            if self.humai_collection_id >= 0:
                print("[_create_fields_and_collection] Found collection '{}', ID: {}".format(humai_pred_collection, self.humai_collection_id))
                                       
            success = self.humai_database_id >= 0 and self.humai_collection_id >= 0
            
            if success:
                
                print("Checking fields...")
                success = success & self._create_elements('database/{}/fields'.format(self.humai_database_id), df_fields)
                print("Checking interests...")
                success = success & self._create_elements('database/{}/interests'.format(self.humai_database_id), df_interests)
                print("Checking collection fields...")
                success = success & self._create_elements('collection/{}/fields'.format(self.humai_collection_id), df_collection_fields)

                self.ready = success
            
            print("[_create_fields_and_collection] Ready => {}".format(self.ready))
        else:
            
            print("[_create_fields_and_collection] Database {} does not exist.".format(database))
            
            
    def push_profile_data(self, df, mapping, interest_mapping=None, cid_column='Customer_ID', test=False):
        
        if not self.ready:
            print("[push_predictions_to_copernica] Not ready, call create_or_init()")
            return
        
        nn = len(df)
        ii = 0

        dfg = df.groupby(by='Copernica_ID')
        t0 = timer()
        for copernica_id, df_profile in dfg:
            
            copernica_id = int(copernica_id)
        
            if ii % 500 == 0:
                t1 = timer()
                print("{:,}/{:,} ({:.2f} %) elapsed: {}".format(ii, nn, 100.0*ii/nn, str(datetime.timedelta(seconds=(t1-t0)))))

            if copernica_id == -1:
                print("  Skipping invalid Copernica ID: {} with profile {}".format(copernica_id, df_profile))
                ii = len(df_profile) + ii
            else:
                profile = df_profile.iloc[0,:]

                #response, data = self.api.get('profile/{}'.format(copernica_id), verbose=False)
                #if response.status_code == 200:#requests.codes.ok:
                    
                fields_data = {
                    key : profile[ mapping[key] ] for key in mapping.keys()
                }

                interests_data = None
                if interest_mapping is not None:
                    interests_data = {
                        key : profile[ interest_mapping[key] ] for key in interest_mapping.keys()
                    }

                if test:
                    print("Copernica ID {}, push:".format(copernica_id))
                    for field in fields_data.keys():
                        print("  '{}' -> '{}'".format(field, fields_data[field]))

                    if interests_data is not None:
                        for interest in interests_data.keys():
                            print("  interest '{}' -> '{}'".format(interest, interests_data[interest]))
                else:
                    response = self.api.put('profile/{}/fields'.format(copernica_id), data=fields_data, verbose=False)
                    if interests_data is not None:
                        response = self.api.put('profile/{}/interests'.format(copernica_id), data=interests_data, verbose=False)

                #else:
                #    print("Could not retrieve Copernica ID {}".format(copernica_id))

            ii = len(df_profile) + ii
                      
        t1 = timer()
        print("{:,}/{:,} ({:.2f} %) elapsed: {}".format(
            nn, nn, 100.0, 
            str(datetime.timedelta(seconds=(t1-t0)))
        ))
        
        
    def push_subprofile_data(self, df, mapping, cid_column='Customer_ID', test=False):
    
        if not self.ready:
            print("[push_predictions_to_copernica] Not ready, call create_or_init()")
            return
    
        nn = len(df)
        ii = 0

        dfg = df.groupby(by='Copernica_ID')
        t0 = timer()
        for copernica_id, df_subprofiles in dfg:
            
            copernica_id = int(copernica_id)
            
            if copernica_id == -1:
                print("  Skipping invalid Copernica ID: {} with profile {}".format(copernica_id, df_subprofiles))
                ii = len(df_profile) + ii
            else:
                #response, data = self.api.get('profile/{}'.format(copernica_id), verbose=False)
                #if response.status_code == 200:#requests.codes.ok:

                #
                # Create predictions as subprofiles
                #

                for index, row in df_subprofiles.iterrows():

                    if ii % 500 == 0:
                        t1 = timer()
                        print("{:,}/{:,} ({:.2f} %) elapsed: {}".format(ii, nn, 100.0*ii/nn, str(datetime.timedelta(seconds=(t1-t0)))))

                    subprofile_data = {
                        key : row[ mapping[key] ] for key in mapping.keys()
                    }

                    if test:
                        print("Copernica ID {}:".format(copernica_id))
                        for field in subprofile_data.keys():
                            print('  {} -> {}'.format(field, subprofile_data[field]))
                    else:                            
                        response = self.api.post(
                            'profile/{}/subprofiles/{}'.format(copernica_id, self.humai_collection_id), 
                            data=subprofile_data, verbose=False
                        )

                    ii = 1 + ii

                    #response.status_code, response.text, response.reason
                #else:
                #    print("Could not retrieve Copernica ID {}".format(copernica_id))

                #    ii = len(df_predictions) + ii            

        t1 = timer()
        print("{:,}/{:,} ({:.2f} %) elapsed: {}".format(
            nn, nn, 100.0, 
            str(datetime.timedelta(seconds=(t1-t0)))
        ))

            
    def push(self, df, pred_type='PersonalizedMailFlow', segment_columns = None, segment_names = None, verbose=False):

        assert False, "Deprecated"
        
        if not self.ready:
            print("[push_predictions_to_copernica] Not ready, call create_or_init()")
            return
        
        nn = len(df)
        ii = 0

        dfg = df.groupby(by='Copernica_ID')
        for copernica_id, df_predictions in dfg:
            
            if copernica_id == -1:
                print("  Skipping invalid Copernica ID: {} with predictions {}".format(copernica_id, df_predictions))
                ii = len(df_predictions) + ii
            else:
                customer_id = df_predictions.iloc[0, :]['Customer_ID']

                response, data = self.api.get('profile/{}'.format(copernica_id), verbose=False)
                if response.status_code == 200:#requests.codes.ok:

                    if verbose:
                        print("Copernica ID {}, Humai Customer ID {} == {}".format(data['ID'], data['fields']['Humai_Customer_ID'], customer_id))
                        for field in data['fields']:
                            print("  {} -> {}".format(field, data['fields'][field]))

                    #
                    # Update existing profile fields
                    #

                    fields_data = {
                        'Humai_Customer_ID' : str(customer_id)
                    }
                    response = self.api.put('profile/{}/fields'.format(copernica_id), data=fields_data, verbose=False)

                    if segment_columns is not None:
                        interests_data = {}
                        row = df_predictions.iloc[0,:]
                        for col,name in zip(segment_columns, segment_names):
                            interests_data[name] = row[col]

                        # Update (PUT) existing interests
                        response = self.api.put('profile/{}/interests'.format(copernica_id), data=interests_data, verbose=False)

                    #
                    # Create predictions as subprofiles
                    #

                    for index, row in df_predictions.iterrows():

                        if ii % 500 == 0:
                            print("{:,}/{:,} ({:.2f} %)".format(ii, nn, 100.0*ii/nn))

                        subprofile = {
                            'Humai_Customer_ID' : str(customer_id),
                            'Humai_Mail_ID' : int(row['Email_ID']),
                            'Type' : pred_type,
                            'Date' : row['Date'].strftime('%Y-%m-%d'),
                            'DatetimeSlot' : row['Time_Of_Day'],
                            'Channel' : 'Email',
                            'Group' : 'B'
                        }

                        response = self.api.post(
                            'profile/{}/subprofiles/{}'.format(copernica_id, self.humai_collection_id), 
                            data=subprofile, verbose=False
                        )

                        ii = 1 + ii

                        #response.status_code, response.text, response.reason
                else:
                    print("Could not retrieve Copernica ID {}".format(copernica_id))

                    ii = len(df_predictions) + ii
            
            
        print("{:,}/{:,} ({:.2f} %)".format(nn, nn, 100.0))
        
    
    def remove_all_subprofiles(self, page_size=1000, verbose=False):
        
        if not self.ready:
            print("[remove_all_subprofiles] Not ready, call create_or_init()")
            return
        
        cur_profile = 0
        total_profiles = 1
        n_removed = 0
        while cur_profile < total_profiles:

            parameters = {
                'start' : cur_profile,
                'limit' : page_size
            }
            response, data = self.api.get(
                'database/{}/profiles'.format(self.humai_database_id), 
                params=parameters, 
                verbose=False
            )
            if 200 <= response.status_code <= 299:

                total_profiles = data['total']

                for profile in data['data']:

                    if cur_profile % 1000 == 0:
                        print("{:,}/{:,} ({:.2f}%). Removed {:,} subprofiles so far".format(
                            cur_profile, total_profiles, 
                            100.0*cur_profile / total_profiles, 
                            n_removed,
                        ))

                    copernica_id = int( profile['ID'] )

                    response, data = self.api.get('profile/{}/subprofiles/{}'.format(copernica_id, self.humai_collection_id), verbose=False)
                    if 200 <= response.status_code <= 299:#requests.codes.ok:

                        for subprofile in data['data']:
                            n_removed = 1 + n_removed
                            self.api.delete('subprofile/{}'.format(subprofile['ID']))
                            if verbose:
                                print("  Deleted subprofile ID {}, created {}, type {}".format(subprofile['ID'], subprofile['created'], subprofile['fields']['Type']))

                    cur_profile = 1 + cur_profile
                    
            else:
                print("[link_and_add_to_customers] Call to retrieve {} profiles starting from {} failed.".format(total_profiles, cur_profile))


    def remove(self, df, pred_type = 'PersonalizedMailFlow', verbose=False):

        #assert False, "Deprecated"
        
        if not self.ready:
            print("[push_predictions_to_copernica] Not ready, call create_or_init()")
            return
        
        cpids = df['Copernica_ID'].unique().tolist()
        nn = len(cpids)
        ii = 0
        
        for copernica_id in cpids:
            
            if copernica_id == -1:
                print("  Skipping invalid Copernica ID: {}".format(copernica_id))
            
            if ii % 500 == 0:
                print("{:,}/{:,} ({:.2f} %)".format(ii, nn, 100.0*ii/nn))
            
            if verbose:
                print("Copernica ID {}".format(copernica_id))

            response, data = self.api.get('profile/{}/subprofiles/{}'.format(copernica_id, self.humai_collection_id), verbose=False)
            if response.status_code == 200:#requests.codes.ok:

                for subprofile in data['data']:
                    if subprofile['fields']['Type'] == pred_type:
                        self.api.delete('subprofile/{}'.format(subprofile['ID']))
                        if verbose:
                            print("  Deleted subprofile ID {}, created {}, type {}".format(subprofile['ID'], subprofile['created'], subprofile['fields']['Type']))
                        
            ii = 1 + ii
            
        print("{:,}/{:,} ({:.2f} %)".format(nn, nn, 100.0))
        

def get_database_id(api, database):
    """
    Get the 'ID' field for database in Copernica.
    """
    
    response, data = api.get('databases')
    if response.status_code == 200:
            
        databases = {db['name']:db['ID'] for db in data['data']}
        return int(databases[database]) if database in databases else -1
                
    return -1


class CopernicaCustomersExtract:
    """
    Helper class for converting between Customer IDs and Copernica IDs, given a customers table.
    """
    
    def __init__(self, gbq, access_token):
        
        self.df_customers = None
        self.gbq = gbq
        self.api = CopernicaREST(access_token)
        
        return
    
    def _extract_string(self, record, field_name):
        
        return record[field_name].strip().upper() if field_name in record else ''
    
    def _extract_int(self, record, field_name):
        
        return int(record[field_name]) if field_name in record else np.nan
    
    def extract(self, database='', page_size=1000):
        
        if self.api is None:
            print("[CopernicaCustomersExtract::extract] No API.")
            return None

        database_id = get_database_id(self.api, database)
        if database_id < 0:
            if verbose:
                print("[CopernicaCustomersExtract::extract] Failed to find database {}".format(database))
            return None
        
        customer_rows = []
        
        cur_profile = 0
        total_profiles = 1
        while cur_profile < total_profiles:

            parameters = {
                'start' : cur_profile,
                'limit' : page_size
            }
            response, data = self.api.get('database/{}/profiles'.format(database_id), params=parameters, verbose=False)
            if 200 <= response.status_code <= 299:

                total_profiles = data['total']

                for profile in data['data']:

                    if cur_profile % 1000 == 0:
                        print("{:,}/{:,} ({:.2f}%)".format(
                            cur_profile, total_profiles, 100.0*float(cur_profile)/total_profiles
                        ))

                    copernica_id = -1
                    if 'fields' in profile:
                        fields = profile['fields']
                        #print("fields:\n============")
                        #print(fields)
                        cfirstname = self._extract_string(fields, 'MFirstName')
                        clastname = self._extract_string(fields, 'MLastName')
                        cemail = self._extract_string(fields, 'MEmail')
                        copernica_id = int( profile['ID'] )

                        unsub = profile['removed'] or self._extract_string(fields, 'MSubscriberRawState') != 'STATUS_SUBSCRIBED' or self._extract_string(fields, 'MNewsletter') != 'YES' or self._extract_string(fields, 'AfgemeldNieuwsbrief') == 'JA'

                        customer_rows.append((
                            copernica_id,cemail,
                            cfirstname,clastname,
                            self._extract_string(fields, 'MDOB'),
                            self._extract_string(fields, 'MGender'),
                            not unsub
                        ))
                        
                    cur_profile = 1 + cur_profile
            else:
                cur_profile = page_size + cur_profile
                print("[link_and_add_to_customers] Call to retrieve {} profiles starting from {} failed.".format(total_profiles, cur_profile))

        print("{:,}/{:,} ({:.2f}%)".format(
            total_profiles, total_profiles, 100.0, 
        ))

        df_customers = pd.DataFrame(
            customer_rows,
            columns=['Copernica_ID','Email','Firstname','Lastname',
                     'Date_of_Birth',
                     'Gender',
                     'Subscribed']
        )

        return df_customers
    
     